#include <stdio.h>
#include "uexec.h"
#include "uart.h"

unsigned char task_one_stack[200];
unsigned char task_one_msgQ [100];
int taskOneTid;

unsigned char task_two_stack[200];
unsigned char task_two_msgQ [100];
int taskTwoTid;

unsigned char task_3_stack[200];
int task3Tid;

void One(void);
void Two(void);
void RunsOffClock(void);
unsigned int msec(unsigned int time);
//
// main
//

int main(void)
{ 

   UART_Init();
   
   UPRINT("\r\n\r\n-------------------- main is alive--------------------\r\n"); 
    
   *LED = (U8)~0xff;
   
   taskOneTid = UEXC_CreateTask(
                     One,                          // entry point
                     task_one_stack,                // stack ptr
                     sizeof(task_one_stack),        // stack size
                     task_one_msgQ,                 // msgQ ptr
                     sizeof(task_one_msgQ),         // msqQ size
                     (U16)UEXC_DEFAULT_TIME_SLICE,  // time slice mills
                     100);                          // task's priority

   taskTwoTid  = UEXC_CreateTask(
                     Two,                           // entry point
                     task_two_stack,                // stack ptr
                     sizeof(task_two_stack),        // stack size
                     task_two_msgQ,                 // msgQ ptr
                     sizeof(task_two_msgQ),         // msqQ size
                     (U16)UEXC_DEFAULT_TIME_SLICE,  // time slice mills
                     (U8)100);                      // task's priority (higher)

   task3Tid    = UEXC_CreateTask(
                     RunsOffClock,                  // entry point
                     task_3_stack,                  // stack ptr
                     sizeof(task_3_stack),          // stack size
                     0,                             // msgQ ptr  (none)
                     0,                             // msqQ size
                     (U16)UEXC_DEFAULT_TIME_SLICE,  // time slice mills
                     100);                          // task's priority (higher)
      
	UEXC_StartScheduler(); // won't return  

   UPRINT("\r\nmain still alive, can't happen");
	UPRINT("\r\nmain is finished");
   
   return 0;
}


//
// task One
//
extern void One(void)
{
   unsigned char msgBuf[20];
   U8  xmtTid, msgSize, i;
   U16 tmoMills=3000;
   
   while (1)
   {
     UPRINT("\r\n\r\ntask One entering 1 sec spin loop");
     msec(1000);
     
     if (UEXC_ReadMsg( msgBuf, sizeof(msgBuf), &xmtTid, &msgSize, tmoMills))
     {
        UPRINT("\r\ntask One got msg of 0x"); UPRINTU8(msgSize);
        UPRINT(" bytes from task id 0x");  UPRINTU8(xmtTid);
        UPRINT("\r\n");
        for (i=0; i<msgSize; ++i)
        {
           UPRINT(" "); UPRINTU8(msgBuf[i]);
        }
     }
     else 
        UPRINT("\r\ntask One UEXC_ReadMsg failed");
   }
}

//
// task Two
//
extern void Two(void)
{
   U8  msgBuf[]={0,1,2,3,4,5,6,7,8,9};


	while (1)
   {
     UPRINT("\r\n\r\ntask Two 1 sec spin loop");
     msec(1000);

     
     UPRINT("\r\ntask Two calling UEXC_SendMsg");
     if (UEXC_SendMsg(msgBuf, sizeof(msgBuf), taskTwoTid, taskOneTid))
     {
        UPRINT("\r\ntask Two sent msg of 0x"); 
        UPRINTU8(sizeof(msgBuf)); UPRINT(" bytes\r\n");
        //for (i=0; i<sizeof(msgBuf); ++i)
        //{
        //   UPRINT(" "); UPRINTU8(msgBuf[i]);
        //}
     }
     else 
        UPRINT("\r\ntask Two UEXC_SendMsg failed");

     UPRINT("\r\ntask Two calling UEXC_Sleep 1 sec");    
     UEXC_Sleep(1,0);                           

     ++msgBuf[0];
   }
}


extern void RunsOffClock(void)
{
   
   unsigned int ticks;
   
   UEXC_Attach(2,0); // attach to clock

   while (1)
   {
      //UPRINT("\r\ntask RunsOffClock calling UEXC_Sleep UEXC_SLEEP_ATTACHED");    
      
      UEXC_Sleep( UEXC_SLEEP_ATTACHED ,0);  // sleep until run off clock

      ticks = UEXC_GetClockTicks();         // get current clock ticks

      UPRINT("\r\nRunsOffClock ticks 0x"); UPRINTU16(ticks);

   }

}
