#include <io.h>
#include <interrupt.h>
#include <string.h>
#include "signal.h"
#include "_uexec.h"
#include "uexec.h"
#include "uart.h"
#include "uartIntr.h"
                                           

//
// circular receive queue (len,data,len....len,data,len)
//   
// note** each entry has its length before and after it so the
//     ** line editing commands can go backwards in the queue
//         
   U8   IUART_RecvQue[MAX_UART_RECV_CHARS];
   U8*  IUART_RecvGetPtr;
   U8*  IUART_RecvPutPtr;
   U8*  IUART_Recv1stPtr;
   U8*  IUART_RecvEndPtr;
   U8   IUART_numLinesRcvd;    // # of complete lines in queue
   U8   IUART_RecvSigReqTid;   // tid wf next line rcvd (0xff = none) 

   U8   IUART_EditMode;        // input is being edited
   U8*  IUART_RecvLinePtr;     // pointer to line being typed
   U8*  IUART_PrevLinePtr;     // pointer to a previous line  
   U8   IUART_RecvLineCnt;     // # chars in line being typed
   U8   IUART_RecvLineCursor;  // offset to cursor of line being typed (0-n)
   U8   IUART_PrevChar;        // previous char typed
   U8   IUART_NumDisplayed;    // # of chars displayed

                               // buffer of unedited line chars
   U8   IUART_UndoLineBuf[IUART_MAX_UNDO_CHARS];

                               // edit line back space buffer
   U8   tmpLine[IUART_MAX_LINE_CHARS*3];

                               // for unEraseInputLine
   U8   old_IUART_RecvLineCnt;   
   U8   old_IUART_RecvLineCursor;
   U8*  old_IUART_RecvPutPtr;    
   U8   old_IUART_EditMode;  
                               // 'getchar' buffer
   U8   IUART_RecvGetChrBuf[IUART_MAX_LINE_CHARS];    
   U8   IUART_RecvGetChrCnt;   // 'getchar' count 
   U8   IUART_RecvGetChrIdx;   // 'getchar' buffer index

//
// circular send queue    (len,data....len,data)
//         
   U8   IUART_SendQue[MAX_UART_SEND_CHARS];
   U8*  IUART_SendGetPtr;
   U8*  IUART_SendPutPtr;
   U8*  IUART_Send1stPtr;
   U8*  IUART_SendEndPtr;
   U8*  IUART_SendLinePtr;     // pointer to count of line last xmited
   U8   IUART_SendLineTot;     // total chars to xmit in line 
   U8   IUART_SendLineCnt;     // number chars xmited in line
   U8   IUART_numLines2send;
   U8   IUART_SendSigReqTid;   // tid wf print completed (0xff = none)

                               // last user line xmited
   U8   prompt[IUART_MAX_LINE_CHARS];

   U8   traceIdx;              
   U8   traceBuf[MAX_TRACE_BUF];

/* end-of-line string = 'Line End' + 'Line Feed' character */
U8 IUART_pszEndOfLine[3] = {0x0d,0x0a,0};

extern TaskControlBlock* current_task;   // uexec's current tcb ptr

char unqueLineRcvd(U8* buffer_p, U8 bufLen, U8* numBytesRcvd);
char queLine2xmt(U8* buffer_p, U8 numBytesToSend);
void dumpQueue(U8* que1st, U8* queEnd, U8* fromPtr, U8* toPtr);
void IUART_SendByte(U8 Data);             
char isEditChar(U8* charRcvd);
void IUART_sPrintfU4(U8 Data, U8* Str);
U8 incXmtGetPtr(void);
U8 incXmtPutPtr(void);
U8 incRcvGetPtr(void);
U8 incRcvPutPtr(void);
void IUART_SendNextChar(void);
void editInputLine(U8 charRcvd);
void rePrintInputLine(U8 onNewLine);
void getPrevInputLine(void);
void getNextInputLine(void);
U8* decQuePtr(U8* ptrGet, U8* ptr1st, U8* ptrEnd);
U8* incQuePtr(U8* ptrGet, U8* ptr1st, U8* ptrEnd);
void delCharLeftOfCursor(void);
void delCharRightOfCursor(void);
void moveCursorLeft(void);
void moveCursorRight(void);
void displayEditHelp(void);
void undoInputLineEdits(void);
void saveInputLine(void);
void moveCursorHome(void);
void moveCursorEol(void);
void eraseInputLine(void);
void insertInputChar(U8 newChr);
void rePrintPrompt(void);
void dumpRcvQue(void);
void dumpXmtQue(void);
char isEndOfLine(U8* chr);
void eraseDisplayedChars(void);
void unEraseInputLine(void);
void dumpTraceBuf(void);

//
// initialize interrupt driven UART handler
//
void IUART_Init(void)
{
   
   DISABLE_RX_TX_INTR;  
                                // Enable External SRAM
   outp((1<<SRE), MCUCR);         // no wait states 

   IUART_RecvGetPtr     = &IUART_RecvQue[0];
   IUART_RecvPutPtr     = &IUART_RecvQue[0];
   IUART_Recv1stPtr     = &IUART_RecvQue[0];
   IUART_RecvLinePtr    = &IUART_RecvQue[0];
   IUART_PrevLinePtr    = &IUART_RecvQue[0];
   IUART_RecvEndPtr     = &IUART_RecvQue[MAX_UART_RECV_CHARS];
   IUART_RecvGetChrCnt  = 0;
   IUART_numLinesRcvd   = 0;
   IUART_RecvSigReqTid  = NO_SIGNAL_PENDING;
   IUART_EditMode       = 0;
   IUART_RecvLineCnt    = 0;
   IUART_RecvLineCursor = 0;
   IUART_PrevChar       = 0;
   IUART_NumDisplayed   = 0;

   IUART_SendGetPtr     = &IUART_SendQue[0];
   IUART_SendPutPtr     = &IUART_SendQue[0];
   IUART_Send1stPtr     = &IUART_SendQue[0];
   IUART_SendEndPtr     = &IUART_SendQue[MAX_UART_SEND_CHARS];
   IUART_numLines2send  = 0;
   IUART_SendSigReqTid  = NO_SIGNAL_PENDING;
   IUART_SendLineTot    = 0;
   IUART_SendLineCnt    = 0;

                                       // set baud rate 
   outp( (U8)IUART_BAUD_SELECT, UBRR);

   ENABLE_RX_TX_INTR;

}
    
//--------------------------------------------------------------------
// UART xmit interrupt handler
//
SIGNAL(SIG_UART_TRANS)                       
{
   
   //DPRINT("\r\nSIG_UART_TRANS IUART_SendGetPtr 0x");
   //DPRINTU16((U16)IUART_SendGetPtr);
   //DPRINT(" ,IUART_SendPutPtr 0x");
   //DPRINTU16((U16)IUART_SendPutPtr);

                                         // queue empty?
   if (IUART_numLines2send > 0)
   {
                                         // send next character
      IUART_SendNextChar();
   }
   else
   {
                                        // disable TX uart interrupts
      outp((1<<RXCIE)|(1<<RXEN)|(1<<TXEN), UCR);

      //DPRINT("\r\nnuartXmtIntrHandler got intr, nothing to send");
      //DPRINT(" ,IUART_SendPutPtr 0x"); DPRINTU16((U16)IUART_SendPutPtr);
      //DPRINT(" ,IUART_SendGetPtr 0x"); DPRINTU16((U16)IUART_SendGetPtr);
      //DPRINT(" ,IUART_SendLineCnt 0x"); DPRINTU16((U16)IUART_SendLineCnt);
      //DPRINT(" ,IUART_numLines2send 0x"); DPRINTU16((U16)IUART_numLines2send);
   }
}

//
// send next character
//
void IUART_SendNextChar(void)
{


   //DPRINT("\r\nIUART_SendNextChar - IUART_numLines2send 0x");
   //DPRINTU8(IUART_numLines2send); 
   //DPRINT(",  IUART_SendLineCnt 0x"); DPRINTU8(IUART_SendLineCnt); 
   //DPRINT("\r\n");

   //
   // finished sending a line
   //
   if (IUART_SendLineCnt == IUART_SendLineTot && IUART_SendLineTot > 0)
   {                               
      //DPRINT("\r\nIUART_SendNextChar IUART_SendSigReqTid 0x");
      //DPRINTU8(IUART_SendSigReqTid);

                                      // does task want know line printed?
      if (IUART_SendSigReqTid != NO_SIGNAL_PENDING)
      {
         //DPRINT("\r\nIUART_SendNextChar sent signal to tid 0x");
         //DPRINTU8(IUART_SendSigReqTid);
         UEXC_SendSignal( LINE_XMTD_SIGNAL, IUART_SendSigReqTid);
         IUART_SendSigReqTid = NO_SIGNAL_PENDING;
      }

      IUART_SendLineCnt = 0;
      IUART_SendLineTot = 0;
      --IUART_numLines2send;
   }

   //
   // setup for new line
   //
   if (IUART_SendLineCnt == 0 && IUART_numLines2send > 0)     
   {

      //DPRINT("\r\nIUART_SendNextChar - IUART_numLines2send 0x");
      //DPRINTU8(IUART_numLines2send); 
                                          // pointer to line xmiting
      IUART_SendLinePtr = IUART_SendGetPtr;
                                          // # chars in line being xmited
      IUART_SendLineTot = *IUART_SendGetPtr; 

      //DPRINT(",  IUART_SendLineCnt 0x"); DPRINTU8(IUART_SendLineCnt); 
      //DPRINT("\r\n");

      if (incXmtGetPtr() == 0)
      {
         DPRINT("\r\nIUART_SendNextChar - IUART_SendQue corrupt");
         dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
                   IUART_SendGetPtr, IUART_SendPutPtr);
      }
                                          
      //dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
      //          IUART_SendGetPtr, IUART_SendPutPtr);
                                          
   }

   //
   // send next character of line
   //
   if (IUART_SendLineCnt < IUART_SendLineTot)     
   {
       //DPRINT("\r\nIUART_SendNextChar IUART_SendGetPtr 0x");
       //DPRINTU16((U16)IUART_SendGetPtr);
       //DPRINT(", *IUART_SendGetPtr 0x"); DPRINTU8(*IUART_SendGetPtr);
   
       IUART_SendByte(*IUART_SendGetPtr);

       ++IUART_SendLineCnt;

       if (incXmtGetPtr() == 0)
       {
          DPRINT("\r\nIUART_SendNextChar - IUART_SendQue corrupt");
          dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
                    IUART_SendGetPtr, IUART_SendPutPtr);
       } 
   }
   else                                 // disable TX uart interrupts
      outp((1<<RXCIE)|(1<<RXEN)|(1<<TXEN), UCR);
}

//-------------------------------------------------------------------------
// UART Receive Interrupt handler 
//
SIGNAL(SIG_UART_RECV)              
{
   U8 charRcvd;

   charRcvd = inp(UDR);                  // get next char

   //
   //  edit input line
   //
   if (isEditChar(&charRcvd)) 
   {
      if (IUART_PrevChar != IUART_ESC) 
         editInputLine(charRcvd);
   }

   //
   // add char to input line
   //
   else 
   {
                                                  // setup for a new line
      if ( IUART_RecvLineCnt == 0) 
      {
                                     // save pointers to next line
         IUART_RecvLinePtr    = IUART_RecvPutPtr;
         IUART_PrevLinePtr    = IUART_RecvLinePtr;

         if (incRcvPutPtr() == 0) // make room for count of next line
         {
            //DPRINT("\r\nnuartRcvIntrHandler - IUART_RcvQue full");
            //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
            //          IUART_RecvGetPtr, IUART_RecvPutPtr);
         }
   
         IUART_RecvLineCursor = 0;
         IUART_NumDisplayed   = 0;
         IUART_EditMode       = 0;
      }
                                                  // add to current line
      if (isEndOfLine(&charRcvd) || IUART_EditMode == 0)              
      {
         *IUART_RecvPutPtr = charRcvd; // append char rcvd        
         ++IUART_RecvLineCnt;
         ++IUART_NumDisplayed;
         ++IUART_RecvLineCursor;
         #ifdef IUART_ECHO_INPUT               
         queLine2xmt(&charRcvd, 1);    // echo char rcvd
         #endif
      }
      else
      {
         insertInputChar(charRcvd);   // insert char rcvd
         ++IUART_RecvLineCnt;
         ++IUART_RecvLineCursor;
                                      // xmt new input line
         rePrintInputLine(IUART_ON_SAME_LINE);
      }
                                      // clear 1st char len
      *IUART_RecvLinePtr = 0;

                                      // bump put ptr
      if (incRcvPutPtr() == 0)
      {
         //DPRINT("\r\nnuartRcvIntrHandler - IUART_RcvQue full");
         //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
         //          IUART_RecvGetPtr, IUART_RecvPutPtr);
      }

                                            // end of line? 
      if (isEndOfLine(&charRcvd))
      {
         //DPRINT("\r\nnuartRcvIntrHandler - got end of line");
         //DPRINT(", IUART_RecvLineCnt 0x"); DPRINTU8(IUART_RecvLineCnt);

         *IUART_RecvLinePtr = IUART_RecvLineCnt;  // save 1st char len
         *IUART_RecvPutPtr  = IUART_RecvLineCnt;  // save 2nd char len

         if (incRcvPutPtr() == 0) // skip over end count
         {
            //DPRINT("\r\nnuartRcvIntrHandler - IUART_RcvQue full");
            //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
            //          IUART_RecvGetPtr, IUART_RecvPutPtr);
         }
                                  // tell em we rcvd a line
         ++IUART_numLinesRcvd;
                                  // clear out line status
         IUART_RecvLineCnt    = 0;
         IUART_RecvLineCursor = 0;
         IUART_NumDisplayed   = 0;
         IUART_EditMode       = 0;
                                     // save pointers to next line
         IUART_RecvLinePtr    = IUART_RecvPutPtr;
         IUART_PrevLinePtr    = IUART_RecvLinePtr;
                                            
         //DPRINT("\r\nnuartRcvIntrHandler IUART_RecvSigReqTid 0x");
         //DPRINTU8(IUART_RecvSigReqTid);

         //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
         //          IUART_RecvGetPtr, IUART_RecvPutPtr);

                                              // any body waiting for a line?
         if (IUART_RecvSigReqTid != NO_SIGNAL_PENDING)
         {
            UEXC_SendSignal( LINE_RCVD_SIGNAL, IUART_RecvSigReqTid);
            IUART_RecvSigReqTid = NO_SIGNAL_PENDING;
         }
      }
   }
}

//
// xmit a char via uart
//
void IUART_SendByte(U8 Data)             
{
   outp( Data, UDR );
   if (traceIdx >= MAX_TRACE_BUF)   
      traceIdx = 0;
   traceBuf[traceIdx++] = Data;
}

//
// print a string using interrupts
//
void IUART_PrintfStr(U8* pBuf)            
{
   queLine2xmt(pBuf, strlen(pBuf));
}

//
// print end of line using interrupts
//
void IUART_PrintfEndOfLine()           
{
   queLine2xmt(IUART_pszEndOfLine, strlen(IUART_pszEndOfLine));
}
    
//-------------------------------------------------------------------
// utility routines with out processor specific I/O
//


//
// 'sprint' a hex digit to users buffer
//
void IUART_sPrintfU4(U8 Data, U8* Str)
{
   U8 Character;
   /* save 4-bit hex value */
   Character = Data & 0x0f;
   if (Character > 9)
       Character += 'A' - 10;
   else
       Character += '0';
   *Str = Character;
   *(Str+1) = 0;
}

//
// 'sprint' two hex digit to users buffer
//
void IUART_sPrintfU8(U8 Data, U8* buf)    
{
    U8  i;
    U8* bufPtr;

    bufPtr = buf;
    for (i=1; i>+0; --i) 
       IUART_sPrintfU4(Data>>(i*4), bufPtr);

}

//
// print two hex digit using interrupts
//
void IUART_PrintfU8(U8 Data)    
{    
   U8  buf[3];

   IUART_sPrintfU8(Data, buf);

   queLine2xmt(buf, strlen(buf));
}

//
// 'sprint' four hex digit to users buffer
//
void IUART_sPrintfU16(U16 Data, U8* buf)     
{
   U8* bufPtr;
   U8  i;

   bufPtr = buf;
   for (i=3; i>+0; --i) 
      IUART_sPrintfU4(Data>>(i*4), bufPtr);
}

//
// print four hex digit using interrupts
//
void IUART_PrintfU16(U16 Data)     
{
   U8  buf[5];

   IUART_sPrintfU16(Data, buf);

   queLine2xmt(buf, strlen(buf));
}

//
// 'sprint' eight hex digit to users buffer
//
void IUART_sPrintfU32(U32 Data, U8* buf)    
{
   U8* bufPtr;
   U8  i;

   bufPtr = buf;
   for (i=7; i>+0; --i) 
      IUART_sPrintfU4(Data>>(i*4), bufPtr);
}


//
// print eight hex digit using interrupts
//
void IUART_PrintfU32(U32 Data)    
{
   U8  buf[9];

   IUART_sPrintfU32(Data, buf);

   queLine2xmt(buf, strlen(buf));
}


//
// pull next line received off queue 
//
char unqueLineRcvd(U8* buffer_p, U8 bufLen, U8* numBytesRcvd)
{
   unsigned char* startGetPtr;
   unsigned char  cnt;
   unsigned char  i;

   //DPRINT("\r\nnunqueLineRcvd IUART_RecvGetPtr 0x"); 
   //DPRINTU16((U16)IUART_RecvGetPtr);
   //DPRINT(", IUART_RecvPutPtr 0x"); 
   //DPRINTU16((U16)IUART_RecvPutPtr);
                                              // givem msg byte length
   *numBytesRcvd = *IUART_RecvGetPtr;
   startGetPtr   = IUART_RecvGetPtr;
   

   //DPRINT("\r\nunqueMsgRcvd  numBytesRcvd 0x"); UART_PrintfU8(*numBytesRcvd);  

   if (incRcvGetPtr() == 0)
   {
      //DPRINT("\r\nunqueLineRcvd  IUART_RecvQue corrupt count , numBytesRcvd 0x");
      //DPRINTU8(*numBytesRcvd);
      *numBytesRcvd = 0;
   }

   cnt = 0;
   for (i=0; i<*numBytesRcvd; ++i)
   {                                          // givem data bytes, if room
      if (i < bufLen)
      {
         buffer_p[i] = *IUART_RecvGetPtr;
         ++cnt;
      }

      //DPRINT("\r\nnunqueLineRcvd IUART_RecvGetPtr 0x"); 
      //DPRINTU16((U16)IUART_RecvGetPtr);
      //DPRINT(" ,buffer_p[");DPRINTU8(i);
      //DPRINT("]: 0x");DPRINTU8(buffer_p[i]);

      if (incRcvGetPtr() == 0)
      {
         //DPRINT("\r\nnunqueLineRcvd - IUART_RecvQue corrupt data, *numBytesRcvd 0x");
         //DPRINTU8(*numBytesRcvd);
         //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
         //          IUART_RecvGetPtr, IUART_RecvPutPtr);
         *numBytesRcvd = 0;
      }
   }
                                              // validate length bytes 
   if (*numBytesRcvd != *IUART_RecvGetPtr)
   {
      //DPRINT("\r\nnunqueLineRcvd - IUART_RecvQue corrupt *IUART_RecvGetPtr 0x");
      //DPRINTU8(*IUART_RecvGetPtr);
      //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
      //          IUART_RecvGetPtr, IUART_RecvPutPtr);
      *numBytesRcvd = 0;
   }
                                              // skip over ending len byte
   if (incRcvGetPtr() == 0)
   {
      //DPRINT("\r\nnunqueLineRcvd - IUART_RecvQue corrupt data, *numBytesRcvd 0x");
      //DPRINTU8(*numBytesRcvd);
      //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
      //          IUART_RecvGetPtr, IUART_RecvPutPtr);
      *numBytesRcvd = 0;
   }
                                              // got entire line
   --IUART_numLinesRcvd;


   //DPRINT("\r\nnunqueLineRcvd - cnt 0x"); DPRINTU8(cnt);
   //DPRINT(", bufLen 0x"); DPRINTU8(bufLen);

   //DPRINT("\r\nnunqueLineRcvd IUART_RecvGetPtr 0x"); 
   //DPRINTU16((U16)IUART_RecvGetPtr);
   //DPRINT(", IUART_RecvPutPtr 0x"); 
   //DPRINTU16((U16)IUART_RecvPutPtr);

   if (cnt > 0 && i < bufLen)                 // make sure it fits in user buf
      return 1;
   else
      return 0; // not enough room in buffer
}


//
// queues a line to print
//
char queLine2xmt( U8* buffer_p, U8 numBytesToSend)
{
   unsigned char* startPutPtr;      
   unsigned char  i;

   if (numBytesToSend == 0)
   {
      //DPRINT("\r\nqueLine2xmt just sent 0 bytes!");   
      return 1;
   }

                                          // add msg size to xmt que
   startPutPtr = IUART_SendPutPtr;
   
   //DPRINT("\r\nqueLine2xmt queueing 0x"); DPRINTU8(numBytesToSend);
   //DPRINT(" bytes, startPutPtr: 0x"); DPRINTU16((U16)startPutPtr);

   *IUART_SendPutPtr = numBytesToSend;
                                          // bump xmt put ptr
   if (incXmtPutPtr() == 0)
   {
      IUART_SendPutPtr = startPutPtr;
      DPRINT("\r\nqueLine2xmt queue full");
      dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
                IUART_SendGetPtr, IUART_SendPutPtr);
      return 0;  // que full
   }
                                          // add line data to xmt que
   for (i=0; i<numBytesToSend; ++i)
   { 
      
      //DPRINT("\r\nqueLine2xmt  IUART_SendPutPtr 0x"); 
      //DPRINTU16((U16)IUART_SendPutPtr);
      //DPRINT(" ,buffer_p["); DPRINTU8(i); DPRINT("] 0x"); 
      //DPRINTU8(buffer_p[i]);

      *IUART_SendPutPtr = buffer_p[i];
                                          // bump xmt put ptr
      if (incXmtPutPtr() == 0)
      {
         IUART_SendPutPtr = startPutPtr;
         DPRINT("\r\nqueLine2xmt queue full");   
         dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
                   IUART_SendGetPtr, IUART_SendPutPtr);
         return 0;  // que full
      }
   }
                                          // tell em we have a line to send
   ++IUART_numLines2send;

                                          // start sending bytes, if needed
   if (startPutPtr == IUART_SendGetPtr)
   {
      outp((1<<RXCIE)|(1<<TXCIE)|(1<<RXEN)|(1<<TXEN), UCR);
      IUART_SendNextChar();
   }
   
   return 1;   
}


//
// debug dump part of any queue
//
void dumpQueue(U8* que1st, U8* queEnd, U8* fromPtr, U8* toPtr)
{
   U8 numBytes;
   
   DPRINT("\r\ndumpQueue fromPtr: 0x"); DPRINTU16((U16)fromPtr);
   DPRINT(" ,toPtr: 0x"); DPRINTU16((U16)toPtr);

   numBytes = 0;
   while (fromPtr != toPtr) 
   {                                       
      DPRINT(" "); DPRINTU8(*fromPtr);  

      ++fromPtr;
      ++numBytes;
      if (fromPtr >= queEnd) 
         fromPtr = que1st;
   }
   DPRINT("\r\n ,numBytes: 0x"); DPRINTU8(numBytes);
}


//
// increment rcv put que pointer
//
U8 incRcvPutPtr(void)
{
   unsigned char* nextPutPtr;
   
   nextPutPtr = IUART_RecvPutPtr + 1;       // try to bump it

   if (nextPutPtr >= IUART_RecvEndPtr) 
     nextPutPtr = IUART_Recv1stPtr;         // wrapped

                                            // check for room  
   if (nextPutPtr == IUART_RecvGetPtr)
   {                                                           
      //DPRINT("\r\nIUART_RecvQue FULL"); 
      return 0;                                 // no room
   }

   IUART_RecvPutPtr = nextPutPtr;           // update ptr
   return 1;
}


//
// increment rcv get que pointer
//
U8 incRcvGetPtr(void)
{
   unsigned char* nextGetPtr;
                                                // check que empty
   if (IUART_RecvGetPtr == IUART_RecvPutPtr)
      return 0;  // nothing to get 
   
   nextGetPtr = IUART_RecvGetPtr + 1;       // try to bump it

   if (nextGetPtr >= IUART_RecvEndPtr ) 
     nextGetPtr = IUART_Recv1stPtr;         // wrapped

   IUART_RecvGetPtr = nextGetPtr;          // update ptr
   return 1;
}


//
// increment xmt put que pointer
//
U8 incXmtPutPtr(void)
{
   unsigned char* nextPutPtr;
   
   nextPutPtr = IUART_SendPutPtr + 1;       // try to bump it

   if (nextPutPtr >= IUART_SendEndPtr) 
      nextPutPtr = IUART_Send1stPtr;        // wrapped

                                                // check for room  
   if (nextPutPtr == IUART_SendGetPtr)
   {                                                           
      //DPRINT("\r\nIUART_SendQue FULL"); 
      return 0;                                 // no room
   }

   IUART_SendPutPtr = nextPutPtr;           // update ptr

   return 1;
}


//
// increment xmt get que pointer
//
U8 incXmtGetPtr(void)
{
   unsigned char* nextGetPtr;
                                                // check que empty
   if (IUART_SendGetPtr == IUART_SendPutPtr)
      return 0;  // nothing to get 
   
   nextGetPtr = IUART_SendGetPtr + 1;       // try to bump it

   if (nextGetPtr >= IUART_SendEndPtr ) 
     nextGetPtr = IUART_Send1stPtr;         // wrapped

   IUART_SendGetPtr = nextGetPtr;          // update ptr
   return 1;
}

//-------------------------------------------------------------------------
// user interface
//

//
// send a line to UART  waits for line to print (tmo 0 just queues it)
//
//    returns 1(ok), 0(fail)
//                                                   
char IUART_Xmt_Line( unsigned char* linePtr,    // ptr to chars to send
                     unsigned char  numChar,    // # of chars to send
                     unsigned int   tmoMills)   // max block wait
{
   unsigned char retVal;
   unsigned char sigVal;
   unsigned char fromTid;
                                         // save last user line xmited
   strncpy(prompt,linePtr,sizeof(prompt));

                                         // no tmo, just queue line
   if (tmoMills == 0)   
      return queLine2xmt(linePtr, numChar);
      
                                         // check uart xmt driver busy 
   if (IUART_SendSigReqTid != NO_SIGNAL_PENDING)
   {
      DPRINT("\r\nIUART_Xmt_Line driver busy, IUART_SendSigReqTid 0x");
      DPRINTU8(IUART_SendSigReqTid);
      return 0; // driver currently busy
   }
                                         // tell driver we want a signal
   IUART_SendSigReqTid = current_task->tid;

   //DPRINT("\r\nIUART_Xmt_Line IUART_SendSigReqTid 0x");
   //DPRINTU8(IUART_SendSigReqTid);
                                         // try to queue chars to be xmitted
   if (queLine2xmt(linePtr, numChar) == 0)
   {
      IUART_SendSigReqTid = NO_SIGNAL_PENDING;
      return 0;
   }
                                         // wait for signal from driver
   retVal = UEXC_WaitSignal( tmoMills, &sigVal, &fromTid);

   //DPRINT("\r\nIUART_Xmt_Line back from UEXC_WaitSignal, retval 0x");
   //DPRINTU8(retVal);
   
   return retVal;
}


//
// recv a line from UART  (null terminated)
//
//    *** task may block if tmoMills is non zero ***
//
//    returns 1(ok), 0(timedout) 
//
char IUART_Rcv_Line( unsigned char* linePtr,    // ptr to chars to be rcvd
                     unsigned char  lineSiz,    // # of chars max in buf
                     unsigned int   tmoMills)   // max mills to wait
{
   static unsigned char retVal;
   static unsigned char sigVal;
   static unsigned char fromTid;
   static unsigned char numRcvd;


   //DPRINT("\r\nIUART_Rcv_Line IUART_numLinesRcvd 0x");
   //DPRINTU8(IUART_numLinesRcvd);

   linePtr[0] = 0; // make a string
                                      // check lines already available
   if (IUART_numLinesRcvd > 0)
   {
      
      //DPRINT("\r\nIUART_Rcv_Line line already avail");
      retVal = unqueLineRcvd(linePtr, lineSiz-1, &numRcvd);
      
      if (retVal == 0)
         DPRINT("\r\nIUART_Rcv_Line unqueLineRcvd failed");
      
      linePtr[numRcvd] = 0; // make a string
      return retVal;
   }

   //DPRINT("\r\nIUART_Rcv_Line tmoMills 0x");
   //DPRINTU8(tmoMills);
                                      // no line is available, can't wait
   if (tmoMills == 0) 
      return 0; // timed out

//                                     
// gotta wait for a signal or tmo
//
                                       // check uart rcv driver busy
   if (IUART_RecvSigReqTid != NO_SIGNAL_PENDING)
   {
      DPRINT("\r\nIUART_Rcv_Line driver busy,  IUART_RecvSigReqTid 0x");
      DPRINTU8(IUART_RecvSigReqTid);
      return 0; // driver currently busy
   }
                                      // tell driver what task is waiting
   IUART_RecvSigReqTid = current_task->tid;

                                      // wait for a signal from driver
   //DPRINT("\r\nIUART_Rcv_Line wait for signal, IUART_RecvSigReqTid 0x");
   //DPRINTU8(IUART_RecvSigReqTid);

   retVal = UEXC_WaitSignal( tmoMills, &sigVal, &fromTid);
   if (retVal == 0) 
      return 0; // timed out


   ///DPRINT("\r\nIUART_Rcv_Line UEXC_WaitSignal IUART_numLinesRcvd 0x");
   ///DPRINTU8(IUART_numLinesRcvd);
   ///DPRINT(" , fromTid 0x"); DPRINTU8(fromTid);
   
   if (IUART_numLinesRcvd > 0 )
   {
      //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
      //          IUART_RecvGetPtr, IUART_RecvPutPtr);

      retVal = unqueLineRcvd(linePtr, lineSiz-1, &numRcvd);
      
      //if (retVal == 0)
      //   DPRINT("\r\nIUART_Rcv_Line unqueLineRcvd failed");
      
      linePtr[numRcvd] = 0; // make a string
      return retVal;
   }
   //else
   //   DPRINT("\r\nIUART_Rcv_Line UEXC_unqueMsgRcvd failed");

   return 0;
}



//
// recv a char from UART  
//
//    *** task may block if tmoMills is non zero ***
//
//    Note: Characters are not available until a line is
//          commited with and end of line character. 
//          Otherwise 'bogus' characters would be returned
//          after line edits were made via Control-C etc.
//
//    returns char(ok), 0(timedout) 
//
char IUART_Rcv_Char( unsigned int tmoMills)   // max mills to wait
{
   U8  charRcvd;
   

   //
   // check char buffer empty
   //
   if (IUART_RecvGetChrCnt == 0) 
   {
      if( IUART_Rcv_Line(IUART_RecvGetChrBuf, sizeof(IUART_RecvGetChrBuf), tmoMills) == 0) 
      {
         DPRINT("\r\ntestIUart IUART_Rcv_Char failed\r\n");
         return 0;
      }
      IUART_RecvGetChrIdx = 0;
      IUART_RecvGetChrCnt = strlen(IUART_RecvGetChrBuf);

      if (IUART_RecvGetChrCnt == 0)
      {
         DPRINT("\r\ntestIUart IUART_Rcv_Char got 0 chars\r\n");
         return 0;
      }

      //DPRINT("\r\ntestIUart IUART_RecvGetChrBuf '"); DPRINT(IUART_RecvGetChrBuf);
      //DPRINT("'\r\n");
   }

   //
   // givem a char
   //

   charRcvd = IUART_RecvGetChrBuf[IUART_RecvGetChrIdx];
   ++IUART_RecvGetChrIdx;
   --IUART_RecvGetChrCnt;
   
   //DPRINT("\r\ntestIUart IUART_Rcv_Char IUART_RecvGetChrCnt 0x");
   //DPRINTU8(IUART_RecvGetChrCnt);
   //DPRINT("\r\n, IUART_RecvGetChrIdx 0x"); DPRINTU8(IUART_RecvGetChrIdx);
   //DPRINT(", charRcvd 0x"); DPRINTU8(charRcvd);
   
   return charRcvd;
}


//--------------------------------------------------------------------------
//  input line editing routines
//

//
// determine if the character rcvd is a control character used to 
// edit the current line being collected (or redisplay it)
//
char isEditChar(U8* charRcvd)
{
   U8 retVal;

   //DPRINT("\r\nisEditChar *charRcvd 0x"); 
   //DPRINTU8(*charRcvd); DPRINT("  ");

//
// check for special keys (arrows, home, and end)   
//
   if (IUART_PrevChar == IUART_ESC)
   {
      if (*charRcvd == '[')     // toss [ after escape
         return 1;

      //DPRINT("\r\nisEditChar IUART_PrevChar 0x"); DPRINTU8(IUART_PrevChar);
      //DPRINT(", *charRcvd 0x"); DPRINTU8(*charRcvd);

      if (*charRcvd == 'A')
         *charRcvd = IUART_CTL_P;  // up arrow
      else if (*charRcvd == 'B')
         *charRcvd = IUART_CTL_N;  // down arrow
      else if (*charRcvd == 'C')
         *charRcvd = IUART_CTL_R;  // right arrow
      else if (*charRcvd == 'D')
         *charRcvd = IUART_CTL_L;  // left arrow
      else if (*charRcvd == 'H')
         *charRcvd = IUART_CTL_B;  // home
      else if (*charRcvd == 'K')
         *charRcvd = IUART_CTL_E;  // end
      else if (*charRcvd == IUART_ESC)
      {
         dumpRcvQue();             // esc
         dumpXmtQue();
         dumpTraceBuf();
      }
   }
      
                                   // save previous char
   IUART_PrevChar  = *charRcvd;

//
//  test for edit char
//
   switch (*charRcvd) 
   {                              
                                  // any of these are edit chars
   case IUART_BS:         
   case IUART_DEL:
   case IUART_ESC:
   case IUART_CTL_B:   
   case IUART_CTL_C:   
   case IUART_CTL_D:   
   case IUART_CTL_E:   
   case IUART_CTL_L:
   case IUART_CTL_P:   
   case IUART_CTL_N:      
   case IUART_CTL_R:
   case IUART_CTL_U:   
      retVal = 1;   
      break;
   default:
      retVal = 0;
   }
   //DPRINT("\r\nisEditChar *charRcvd 0x"); DPRINTU8(*charRcvd);
   //DPRINT(", retVal 0x"); DPRINTU8(retVal);
   return retVal;
}

//
// edit input line
//
void editInputLine(U8 charRcvd)
{
                                        
   switch (charRcvd) 
   {
                               // backspace  (delete one char to left)
   case IUART_BS:
      if (IUART_RecvLineCnt > 0)
         delCharLeftOfCursor();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // delete     (delete one char to right)
   case IUART_DEL:
      if (IUART_RecvLineCnt > 0 && IUART_RecvLineCursor < IUART_RecvLineCnt)
      {
         delCharRightOfCursor();
         rePrintInputLine(IUART_ON_SAME_LINE);
      }
      break;
                               // control-U  (undo all line edits)
   case IUART_CTL_U:
      if (IUART_RecvLineCnt > 0 && IUART_RecvLineCursor < IUART_RecvLineCnt)
      {
         eraseInputLine();
         eraseDisplayedChars();
         undoInputLineEdits();     
         rePrintInputLine(IUART_ON_SAME_LINE);
      }
      break;
                               // control-P  (recall previous line)
   case IUART_CTL_P:
      eraseInputLine();
      eraseDisplayedChars();
      getPrevInputLine();    
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // control-N  (recall next line)
   case IUART_CTL_N:
      eraseInputLine();
      eraseDisplayedChars();
      getNextInputLine();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // control-C  (clear current line)
   case IUART_CTL_C:
      eraseInputLine();
      eraseDisplayedChars();
      break;
                               // control-D  (display current line)
   case IUART_CTL_D:
      rePrintInputLine(IUART_ON_NEW_LINE);
      break;
                               // control-L  (left one char)
   case IUART_CTL_L:        
      moveCursorLeft();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // control-R  (right one char)
   case IUART_CTL_R:
      moveCursorRight();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // control-B  (move cursor to beginning of line)
   case IUART_CTL_B:
      moveCursorHome();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
                               // control-E  (move cursor to end of line)
   case IUART_CTL_E:
      moveCursorEol();
      rePrintInputLine(IUART_ON_SAME_LINE);
      break;
   
   default:
      ;
   }      
}

//
// check if char is an end of line char
//
char isEndOfLine(U8* chr)
{
   switch (*chr)
   {              // these are valid end of line chars
   case ';':
   case '\r':
   case '\n':
      *chr = ';';
      return 1;
   default:
      return 0;
   }
}
//
// reprint partially input line 
//
void rePrintInputLine(U8 onNewLine)
{
   U8  inputLine[IUART_MAX_LINE_CHARS];
   U8* textPtr;
   U8  i;

   //DPRINT("\r\nrePrintInputLine onNewLine 0x"); DPRINTU8(onNewLine);
   //DPRINT(", IUART_RecvLineCnt 0x"); DPRINTU8(IUART_RecvLineCnt);
   //DPRINT(", IUART_EditMode 0x"); DPRINTU8(IUART_EditMode);
   //DPRINT("\r\n ");
   //dumpRcvQue();
                                    // issue a line feed
   if (onNewLine)
   {
      queLine2xmt("\n",1);
      rePrintPrompt();
   }
   else
   {                                 // erase old chars from display
      eraseDisplayedChars();
   }
   //                                
   // print input so far
   //
   if (IUART_RecvLineCnt == 0 )
      return;
                                    // get pointer to text to display
   textPtr = incQuePtr(IUART_RecvLinePtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   
   if (IUART_EditMode == 0)
   {
      //
      // print unedited input line
      //
      for (i=0; i<IUART_RecvLineCnt; ++i) 
      {
         inputLine[i] = *textPtr;
         textPtr = incQuePtr(textPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
      }
      queLine2xmt(inputLine, IUART_RecvLineCnt);
   }
   else 
   {
      //
      // print edited input line
      //

      if (IUART_RecvLineCursor != 0)
      {                                         // print text before cursor
         for (i=0; i<IUART_RecvLineCursor; ++i) 
         {
            inputLine[i] = *textPtr;
            textPtr = incQuePtr(textPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
         }
         queLine2xmt(inputLine, IUART_RecvLineCursor);
      }
                                                // print cursor
      queLine2xmt(".", 1);

      if (IUART_RecvLineCursor != IUART_RecvLineCnt)
      {                                         // print text after cursor
         for (i=0; i<IUART_RecvLineCnt-IUART_RecvLineCursor; ++i) 
         {
            inputLine[i] = *textPtr;
            textPtr = incQuePtr(textPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
         }
         queLine2xmt( inputLine, IUART_RecvLineCnt-IUART_RecvLineCursor);
      }
   }
                                  // save number of chars displayed
   IUART_NumDisplayed = IUART_RecvLineCnt + IUART_EditMode;
}

//
// clear out chars being displayed
//
void eraseDisplayedChars(void)
{
   U8 numToBlankOut;
   U8 i,j;
                               // get # chars to blank out
   if (IUART_RecvLineCnt + IUART_EditMode < IUART_NumDisplayed)
      numToBlankOut = IUART_NumDisplayed - IUART_RecvLineCnt - IUART_EditMode;
   else
      numToBlankOut = 0;

                                         // validate # of chars to blank out
   if (numToBlankOut > IUART_NumDisplayed) 
   {
      //DPRINT("\r\neraseDisplayedChars invalid numToBlankOut 0x"); DPRINTU8(numToBlankOut);
      numToBlankOut = IUART_NumDisplayed;
   }
                                         // blank em if you got em
   j = 0;
   if (numToBlankOut > 0)
   {                           
      for (i=0; i<numToBlankOut; ++i,++j) 
         tmpLine[j] = IUART_BS; // backup
      for (i=0; i<numToBlankOut; ++i,++j) 
         tmpLine[j] = ' ';      // blank out
   }
                                
   for (i=0; i<IUART_NumDisplayed; ++i,++j) 
      tmpLine[j] = IUART_BS;    // backup

   //DPRINT("\r\neraseDisplayedChars IUART_NumDisplayed 0x"); DPRINTU8(IUART_NumDisplayed);
   //DPRINT(", IUART_RecvLineCnt 0x"); DPRINTU8(IUART_RecvLineCnt);
   //DPRINT("\r\n, IUART_EditMode 0x"); DPRINTU8(IUART_EditMode);
   //DPRINT(", numToBlankOut 0x"); DPRINTU8(numToBlankOut);
   //DPRINT(", j 0x"); DPRINTU8(j);
   //DPRINT("\r\n ");
                                         // display this mess
   queLine2xmt(tmpLine,j);

   IUART_NumDisplayed = 0;
}

//
// reprints last line printed by user
//
void rePrintPrompt(void)
{                                          // give em the prompt again
   queLine2xmt( prompt, strlen(prompt));
}


//
// clear input line
//
void eraseInputLine(void)
{                                          
                                          // save it all
   old_IUART_RecvLineCnt    = IUART_RecvLineCnt;               
   old_IUART_RecvLineCursor = IUART_RecvLineCursor;
   old_IUART_RecvPutPtr     = IUART_RecvPutPtr;
   old_IUART_EditMode       = IUART_EditMode;
                                          // clear it all out
   IUART_RecvLineCnt     = 0;
   IUART_RecvLineCursor  = 0;
   IUART_RecvPutPtr      = IUART_RecvLinePtr;
   IUART_EditMode        = 0;
}

//
// undo eraseInputLine
//
void unEraseInputLine(void)
{                                         // restore it all
   IUART_RecvLineCnt    = old_IUART_RecvLineCnt;               
   IUART_RecvLineCursor = old_IUART_RecvLineCursor;
   IUART_RecvPutPtr     = old_IUART_RecvPutPtr;
   IUART_EditMode       = old_IUART_EditMode;
}

//
// position cursor at beginning of line
//
void moveCursorHome(void)
{
   IUART_RecvLineCursor = 0;
   IUART_EditMode = 1;
}
//
// position cursor at end of line
//
void moveCursorEol(void)
{
   IUART_RecvLineCursor = IUART_RecvLineCnt;
   IUART_EditMode = 1;
}

//
// get previous input line from given pointer to 1st len count of new line
//
//   rcv que format ...len,data,len...len,data,len...
//
void getPrevInputLine(void)
{
   unsigned char* backPtr;
   unsigned char  backLen;
   unsigned char  i;
                                   // make sure we can back up a line
   backPtr = decQuePtr(IUART_PrevLinePtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   backLen = *backPtr;
   
   if (backLen == 0) 
   {
      unEraseInputLine();
      return;
   }

   for (i=0; i<=backLen; ++i) 
      backPtr = decQuePtr(backPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   
   //DPRINT("\r\ngetPrevInputLine backLen 0x"); DPRINTU8(backLen);
   //DPRINT(", *backPtr 0x"); DPRINTU8(*backPtr);
   //DPRINT(", IUART_PrevLinePtr 0x"); DPRINTU16((U16)IUART_PrevLinePtr);

   if (*backPtr != backLen)
   {
      //DPRINT("\r\nngetPrevInputLine lengths don't match que corrupt");
      //dumpRcvQue();
      unEraseInputLine();
      return; // lengths don't match  ??? return current line again ???
   }

                                   // save start of previous input line
   IUART_PrevLinePtr = backPtr;
                                   // drop semicolon on end of line
   --backLen;
                                   // save new count for chars input so far 
   IUART_RecvLineCnt    = backLen;
   IUART_RecvLineCursor = backLen;
                                   // point to begining of input line
   IUART_RecvPutPtr = IUART_RecvLinePtr;

   IUART_EditMode = 0;
                                   // get count and each char  
   for (i=0; i<=backLen; ++i) 
   {
      *IUART_RecvPutPtr = *backPtr;
      backPtr = incQuePtr(backPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
      IUART_RecvPutPtr =
         incQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   }
   
   //dumpRcvQue();
}

//
// get next input line forward from given pointer to 1st len count of new line
//
//   rcv que format ...len,data,len...len,data,len...
//
void getNextInputLine(void)
{
   unsigned char  numNext;
   unsigned char* fwdPtr;
   unsigned char  fwdLen;
   unsigned char  i;
   
                                   // don't move forward past current input line
   if (IUART_PrevLinePtr == IUART_RecvLinePtr)
   {
      unEraseInputLine();
      return;
   }
                                   // make sure we can go forward a line
   fwdPtr = IUART_PrevLinePtr;
   fwdLen = *fwdPtr;
   
   if (fwdLen == 0) 
      return;

   for (i=0; i<=fwdLen; ++i) 
      fwdPtr = incQuePtr(fwdPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   
   //DPRINT("\r\ngetNextInputLine fwdLen 0x"); DPRINTU8(fwdLen);
   //DPRINT(", *fwdPtr 0x"); DPRINTU8(*fwdPtr);
   //DPRINT(", IUART_PrevLinePtr 0x"); DPRINTU16((U16)IUART_PrevLinePtr);

   if (*fwdPtr != fwdLen)
   {
      //DPRINT("\r\nngetnextInputLine lengths don't match que corrupt");
      //dumpRcvQue();
      unEraseInputLine();
      return; // lengths don't match
   }
                                   // point to next lines count
   fwdPtr  = incQuePtr(fwdPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   IUART_PrevLinePtr = fwdPtr;
                                   // get count for next line
   numNext = *IUART_PrevLinePtr;

   //DPRINT("\r\ngetNextInputLine numNext 0x"); DPRINTU8(numNext);
   //DPRINT(", IUART_PrevLinePtr 0x"); DPRINTU16((U16)IUART_PrevLinePtr);

                                   // save new count for chars input so far 
                                    // drop semicolon on end of line
   IUART_RecvLineCnt    = numNext-1;
   IUART_RecvLineCursor = numNext-1;
                                   // point to begining of input line
   IUART_RecvPutPtr = IUART_RecvLinePtr;

                                   // get count and each char
   for (i=0; i<numNext; ++i) 
   {
       *IUART_RecvPutPtr = *fwdPtr;
       IUART_RecvPutPtr =
          incQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
       fwdPtr = incQuePtr(fwdPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   }
                                   
   IUART_EditMode = 0;

   //dumpRcvQue();
}

//
// save input line for undo
//
void saveInputLine(void)
{
   U8* destPtr;
   U8* srcPtr;
   U8  srcCnt;
   U8  i;
                                           // setup ptrs for move
   destPtr  = (U8*)IUART_UndoLineBuf;
   srcPtr   = IUART_RecvLinePtr;
   srcCnt   = IUART_RecvLineCnt;
   *srcPtr  = srcCnt;
                                           // get undo count and chars
   for (i=0; i<=srcCnt; ++i) 
   {
      *destPtr = *srcPtr;
      ++destPtr;
      srcPtr = incQuePtr(srcPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   }

   *destPtr = 0; // make a string
   //DPRINT("\r\nsaveInputLine srcCnt 0x"); DPRINTU8(srcCnt);
   //DPRINT(", IUART_UndoLineBuf '"); DPRINT(&IUART_UndoLineBuf[1]);
   //DPRINT("'\r\n");
}
//
// undo all edits to current input line
//
void undoInputLineEdits(void)
{
   U8* srcPtr;
   U8  srcCnt;
   U8  i;
                                           // setup ptrs for move
   srcPtr   = IUART_UndoLineBuf;
   srcCnt   = IUART_UndoLineBuf[0];
   IUART_RecvPutPtr  = IUART_RecvLinePtr;
                                           //reset input line
   IUART_PrevLinePtr    = IUART_RecvLinePtr;
   IUART_RecvLineCnt    = srcCnt;
   IUART_RecvLineCursor = srcCnt;
                                           // get undo count and chars
   for (i=0; i<=srcCnt; ++i) 
   {
      *IUART_RecvPutPtr = *srcPtr;
      ++srcPtr;
      IUART_RecvPutPtr = 
         incQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   }

   //DPRINT("\r\nundoInputLineEdits srcCnt 0x"); DPRINTU8(srcCnt);
   //DPRINT(", IUART_UndoLineBuf '"); DPRINT(&IUART_UndoLineBuf[1]);
   //DPRINT("'\r\n");
   //dumpRcvQue();
}

//
// decrement que pointer for any queue
//
U8* decQuePtr(U8* ptr, U8* ptr1st, U8* ptrEnd)
{

   --ptr;                    // try to decrement it

   if (ptr < ptr1st) 
     ptr = ptrEnd-1;         // wrapped

   return ptr;
}

//
// increment que pointer for any queue
//
U8* incQuePtr(U8* ptr, U8* ptr1st, U8* ptrEnd)
{

   ++ptr;                    // try to decrement it

   if (ptr >= ptrEnd) 
     ptr = ptr1st;           // wrapped

   return ptr;
}

//
// delete char to left of cursor in input line
//
void delCharLeftOfCursor(void)
{
   U8* ptrCur;
   U8* ptrLast;
   U8  i;
                                       // check for something to delete
   if (IUART_RecvLineCursor == 0) 
      return;
                                       
   //DPRINT("\r\ndelCharLeftOfCursorr IUART_RecvLineCursor 0x"); 
   //DPRINTU8(IUART_RecvLineCursor);
   //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
   //          IUART_RecvGetPtr, IUART_RecvPutPtr);

                                       // point to start of line
   ptrLast = IUART_RecvLinePtr;
                                       // skip count
   //ptrLast  = incQuePtr(ptrLast, IUART_Recv1stPtr, IUART_RecvEndPtr);

                                       // for each char in line
   for (i=0; i<IUART_RecvLineCnt; ++i) 
   {
      ptrCur = incQuePtr(ptrLast, IUART_Recv1stPtr, IUART_RecvEndPtr);
      if (i >= IUART_RecvLineCursor) 
      {
         *ptrLast = *ptrCur;   // shuffle chars left one
         *ptrCur  = 0;
      }
      ptrLast = ptrCur;
   }
                                       // decrement put ptr
   IUART_RecvPutPtr = 
      decQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   
   --IUART_RecvLineCursor;
   --IUART_RecvLineCnt;
                                      // uart intr driver allocates count
   if (IUART_RecvLineCnt == 0) 
      IUART_RecvPutPtr = IUART_RecvLinePtr;

   //dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
   //          IUART_RecvGetPtr, IUART_RecvPutPtr);
}


//
// delete char to right of cursor in input line
//
void delCharRightOfCursor(void)
{
   U8* ptrCur;
   U8* ptrLast;
   U8  i;
                                       // check for something to delete
   if (IUART_RecvLineCursor == IUART_RecvLineCnt) 
      return;
                                       // point to start of line
   ptrLast = IUART_RecvLinePtr;
                                       // skip count
   ptrLast = incQuePtr(ptrLast, IUART_Recv1stPtr, IUART_RecvEndPtr);

                                       // for each char in line
   for (i=0; i<IUART_RecvLineCnt; ++i) 
   {
      ptrCur = incQuePtr(ptrLast, IUART_Recv1stPtr, IUART_RecvEndPtr);
   
                                       // shuffle chars left one 
      if (i >=IUART_RecvLineCursor+1) 
      {
         *ptrLast = *ptrCur;
         *ptrCur  = 0;
      }
      ptrLast = ptrCur;
   }
                                       // decrement put ptr
   IUART_RecvPutPtr = 
      decQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);

   --IUART_RecvLineCnt;
                                       // uart intr driver allocates count
   if (IUART_RecvLineCnt == 0) 
      IUART_RecvPutPtr = IUART_RecvLinePtr;

   //if (IUART_RecvLineCursor == IUART_RecvLineCnt && IUART_EditMode == 1)
   //   IUART_EditMode = 0;
}

//
// insert char typed at cursor 
//
void insertInputChar(U8 newChr)
{
   U8* srcPtr;
   U8* destPtr;
   U8  i;
   
   //DPRINT("\r\ninsertInputChar newChr 0x"); DPRINTU8(newChr);
   //dumpRcvQue();

                                          // setup ptrs for move
   srcPtr  = decQuePtr(IUART_RecvPutPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
   destPtr = IUART_RecvPutPtr;
                                          // shuffle chars right
   for (i=0; i<IUART_RecvLineCnt-IUART_RecvLineCursor; ++i) 
   {
      *destPtr = *srcPtr;
      destPtr = decQuePtr(destPtr, IUART_Recv1stPtr, IUART_RecvEndPtr);
      srcPtr  = decQuePtr(srcPtr,  IUART_Recv1stPtr, IUART_RecvEndPtr);
   }
                                          // insert new char
   *destPtr = newChr;

   //dumpRcvQue();
}

//
// move cursor left one position
//
void moveCursorLeft(void)
{
   if (IUART_RecvLineCursor > 0)
   {
      if (IUART_RecvLineCursor == IUART_RecvLineCnt)
      {
         IUART_EditMode = 1;
         saveInputLine();
      }
      --IUART_RecvLineCursor;
   }
}

//
// move cursor right one position
//
void moveCursorRight(void)
{
   if (IUART_RecvLineCursor < IUART_RecvLineCnt)
   {
      ++IUART_RecvLineCursor;

      //if (IUART_RecvLineCursor == IUART_RecvLineCnt && IUART_EditMode == 1)
      //   IUART_EditMode = 0;
   }
}

//
// debug dump of rcv que
// 
void dumpRcvQue(void)
{
   DPRINT("\r\ndumpRcvQue IUART_RecvGetPtr 0x"); DPRINTU16((U16)IUART_RecvGetPtr);
   DPRINT(", IUART_RecvLinePtr 0x");     DPRINTU16((U16)IUART_RecvLinePtr);
   DPRINT("\r\n, IUART_PrevLinePtr 0x"); DPRINTU16((U16)IUART_PrevLinePtr);
   DPRINT(", IUART_RecvLineCnt 0x");     DPRINTU8(IUART_RecvLineCnt);
   DPRINT(", IUART_RecvLineCursor 0x");  DPRINTU8(IUART_RecvLineCursor);
   DPRINT("\r\n, IUART_EditMode 0x");    DPRINTU8(IUART_EditMode);
   DPRINT(", IUART_NumDisplayed 0x");    DPRINTU8(IUART_NumDisplayed);
   DPRINT(", IUART_RecvSigReqTid 0x");   DPRINTU8(IUART_RecvSigReqTid);
         
   dumpQueue(IUART_Recv1stPtr, IUART_RecvEndPtr,
             IUART_Recv1stPtr, IUART_RecvPutPtr);
   DPRINT("\r\n");
}

//
// debug dump of xmt que
// 
void dumpXmtQue(void)
{
   DPRINT("\r\ndumpXmtQue IUART_SendGetPtr 0x"); DPRINTU16((U16)IUART_SendGetPtr);
   DPRINT(", IUART_SendLinePtr 0x");       DPRINTU16((U16)IUART_SendLinePtr);
   DPRINT("\r\n, IUART_numLines2send 0x"); DPRINTU8(IUART_numLines2send);
   DPRINT(", IUART_SendLineCnt 0x");       DPRINTU8(IUART_SendLineCnt);
   DPRINT(", IUART_SendSigReqTid 0x");     DPRINTU8(IUART_SendSigReqTid);

   dumpQueue(IUART_Send1stPtr, IUART_SendEndPtr,
             IUART_Send1stPtr, IUART_SendPutPtr);

   DPRINT("\r\n");
}

//
// debug dump trace bufer
// 
void dumpTraceBuf(void)
{
   U8  i;  

   DPRINT("\r\ndumpTraceBuf traceIdx 0x"); DPRINTU8(traceIdx);
   DPRINT("\r\n");

   for (i=0; i<MAX_TRACE_BUF; ++i) 
   {
      DPRINT(" "); DPRINTU8(traceBuf[traceIdx++]);
      if (traceIdx >= MAX_TRACE_BUF)
         traceIdx = 0;
   }

   traceIdx = 0;
   DPRINT("\r\n");
}
