#include <stdlib.h>
#include "_uexec.h"
#include "uart.h"
#include <io.h>
#include "signal.h"
#include <interrupt.h>

#ifdef HDSL2
#include "hdsl2api.h"
#include "L1dslusr.h"
#endif

unsigned char numTasks=0;
static TaskControlBlock tcb[NUM_TASKS], *free_list_ptr;
static AttachList_t attachedTasks[NUM_TASKS];

static TaskTrace_t  traceBuf[MAX_TRACE_IDX];
static unsigned int intrTicks=0;
static unsigned int traceIdx =0;
static unsigned int numTraces=0;

Lite_t lites;

#ifdef HDSL2
extern linest_t *cfg_stat2ptr[];
#endif

TaskControlBlock *current_task;
addr_t           uexc_current_sp;   // current uexec task sp
addr_t           uexc_current_func; // current uexec task pc
addr_t           pcCur;             // program counter at timer interrupt
addr_t           spCur;             // stack pointer at timer interrupt
static int next_tid;                // next task id to create
static unsigned char null_task_stack[UEXC_MIN_STACK_SIZE];  

static void InitSystem(void);
void NullTask(void);
void KillTask(int tid);
void justReturns(void);
void UEXC_RunAttachedTask(void);
int  incMsgQuePutPtr(TaskControlBlock* tcbPtr);
int  incMsgQueGetPtr(TaskControlBlock *tcbPtr);
void dumpMsgQue(TaskControlBlock *tcbPtr);
void dumpMsg(TaskControlBlock* tcbPtr, unsigned char* rcvStartPtr, 
             unsigned char* rcvPutPtr);
void dumpStack(TaskControlBlock* tcbPtr, unsigned int spNow);
char UEXC_unqueMsgRcvd( 
                   unsigned char* msgBuf,    // user buffer to recv msg
                   unsigned char  bufSize,   // byte size of buffer
                   unsigned char* xmtTid,    // task id that sent msg
                   unsigned char* msgSize,   // byte size of msg rcvd
                   unsigned int   tmoMills);  // max millisec to wait

void UEXC_DecTaskTimers(void);
void UEXC_DecTimer(void);
void spt_micro_delay(unsigned int time);
unsigned int msec(unsigned int time);
void dumpTcb(TaskControlBlock* tcbPtr);
addr_t getFuncAddr(void (*func) (void));
TaskControlBlock * getHighestPriorTask (void);
void dumpTaskTraces();
void showLites(Lite_t* lites);
#ifdef HDSL2
int unqueEocMsgRcvd(linest_t* irq_lptr,
                    uint8_t* buffer_p, 
                    uint8_t  buffer_sz, 
                    uint8_t* numBytesRcvd);
#endif
/* User Callable  Functions
 * ========================
 */

/* Start the multi-tasking system
 */
void UEXC_StartScheduler(void)
{
   //UPRINT("\r\n\r\nUEXC_StartScheduler calling UEXC_StartTimer");
	UEXC_StartTimer();     

   //UPRINT("\r\nUEXC_StartScheduler calling UEXC_Schedule");
	UEXC_Schedule();  // never comes back
   //while(1)
      //UPRINT("\r\nUEXC_StartScheduler, can't happen");
}

//-------------------------------------------------------------------------------
// timer interrupt handler    GCCAVR
//
//
//   stack looks like this after interrupt stack frame is built
//
//        +37   return address low
//        +36   return address high
//        +35   r31
//        +34   r30
//        +33   r29
//        +31   r28
//        +30   r27
//        +29   r26
//        +28   r25
//        +27   r24
//        +26   r23
//        +25   r22
//        +24   r21
//        +23   r20
//        +22   r19
//        +21   r18
//        +20   r17
//        +19   r16
//        +18   r15
//        +17   r14
//        +16   r13
//        +15   r12
//        +14   r11
//        +13   r10
//        +12   r9
//        +11   r8
//        +10   r7
//        +09   r6
//        +08   r5
//        +07   r4
//        +06   r3
//        +05   r1
//        +04   r0
//        +03   SREG
//        +02   __tmp_reg__
//        +01   __zero_reg__
//    sp->
//            
//        *** frame size depends on compiler  ***
//        *** currently 37 bytes including    ***
//        *** 'reti' return address of intr   ***
//
//        *** any C function that does a RETI ***
//        *** MUST not allocate stack space   ***
//
//        *** grep for Public in uexec.s    ***
//  

SIGNAL(SIG_OUTPUT_COMPARE2)     // signal handler for timer compare 
//SIGNAL(SIG_OUTPUT_COMPARE1A)    // signal handler for timer compare 
{
   static unsigned char spHi, spLo;
   static unsigned int  timerBeat=0;
   
	                                         // save r0-31, SREG, get sp  
   asm volatile(
      "in r0,__SP_L__"   "\n\t"
      "mov %0,r0"        "\n\t"
       : "=r" (spLo)			
       :  
       : "r0","r1","r2","r3","r4","r5","r6","r7","r8", \
         "r9","r10","r11","r12","r13","r14","r15","r16"
   );

   asm volatile(
      "in r17,__SP_H__"   "\n\t"          
      "mov %0,r17"        "\n\t"          
       : "=r" (spHi)			
       :  
       : "r17","r18","r19","r20","r21","r22","r23","r24", \
         "r25","r26","r27","r28","r29","r30","r31"
   );
                                // get stack pointer after all pushes
   spCur = (addr_t)((spHi << 8) | spLo);
                                
                                // get return address for 'reti'
   pcCur = (addr_t)(*((U8*)spCur+UEXC_INTR_STACK_SIZE-1));
   pcCur = (addr_t)(pcCur<<8 | *((U8*)spCur+UEXC_INTR_STACK_SIZE));
   
   //UPRINT("\r\n_TimerInterrupt tid 0x"); UPRINTU8(current_task->tid);
   //UPRINT(" ,spHi 0x");  UPRINTU8(spHi);
   //UPRINT(" ,spLo 0x");  UPRINTU8(spLo);
   //UPRINT(" ,spCur 0x"); UPRINTU16(spCur);
   //UPRINT(" ,pcCur 0x"); UPRINTU16(pcCur);
   //UPRINT(" ,justReturns 0x"); UPRINTU16((addr_t)justReturns);
         
//
// timer led heartbeat *** taking this code out causes GCCAVR not to push r28 ***
//
                                           // shine a little light on the situation
   if (current_task != 0 && current_task->tid < numTasks) 
      lites.tid = current_task->tid;
   
   if (pcCur != (addr_t)justReturns)       // ignore fake interrupts by defer
   {
                                           // flash on/off every second
      ++timerBeat;
      if (timerBeat == 1000/UEXC_CLOCK_TICK_MILLS)    
      {
         lites.tid |= 0x8;  // steal high bit for heartbeat    
      }
      if (timerBeat == 2000/UEXC_CLOCK_TICK_MILLS)
      {
         lites.tid &= 0x7;  // steal high bit for heartbeat    
         timerBeat = 0;
      }      
   }

   #ifdef UEXC_HEART_BEAT
   showLites(&lites);
   #endif
   
//
// task trace buffer   *** taking this code out causes GCCAVR not to push r28 ***
//
   
   if (pcCur != (addr_t)justReturns)          // ignore fake interrupts by defer
   {
      ++intrTicks;
      
      if (traceIdx >= MAX_TRACE_IDX)       // circular trace buffer
         traceIdx = 0;
   
      if (numTraces < MAX_TRACE_IDX)       // count valid traces in buffer
         ++numTraces;
                                           // trace current ready task
      traceBuf[traceIdx].tick = intrTicks;
      traceBuf[traceIdx].pc   = pcCur;
      traceBuf[traceIdx].sp   = spCur;

      if (current_task != 0)
      {
         traceBuf[traceIdx].tid      = current_task->tid;
         traceBuf[traceIdx].state    = current_task->state;
         traceBuf[traceIdx].tmoTicks = current_task->tmoTicks;
      }
      else
      {
         traceBuf[traceIdx].tid   = 0xff;
         traceBuf[traceIdx].state = 0xff;
      }

      ++traceIdx;

   }

   #ifdef TRACE_TASKS  
                                       // dump out task trace history
   if (numTraces == MAX_TRACE_IDX)
   {
      dumpTaskTraces();
      numTraces = 0;    // trace again
   }
   #endif 

//
// make sure we are running our tasks
//
   if (current_task != 0)   
   {
      //                        
      // validate stack pointer is for current task
      //
      if (spCur <= current_task->stack_end   && 
          spCur >= current_task->stack_start &&
          current_task->tid < numTasks)
      {
         //UPRINT("\r\n_TimerInterrupt spCur 0x");                //
         //UPRINTU16((U16)spCur);                                 //
         //UPRINT(" ,pcCur 0x"); UPRINTU16((U16)pcCur);           //
         //UPRINT(" ,uexc_current_func*2 0x");                    //
         //UPRINTU16((U16)uexc_current_func*2);                   //
                                                                     
         //dumpTcb(current_task);                                 //
         //dumpStack(current_task, spCur);                        //
         
                           // check if current task's time slice is over
         UEXC_CheckTask();  // may jump to new task 
         
      }
      else
      {
         if (current_task->tid < numTasks) 
         {
            UPRINT("\r\n_TimerInterrupt invalid stack, tid 0x"); 
            UPRINTU8((U8)current_task->tid);
            #ifdef UEXC_DEBUG                                 
            UPRINT(" spCur 0x"); UPRINTU16((U16)spCur);                                 //
            UPRINT(" ,pcCur 0x"); UPRINTU16((U16)pcCur);           
            UPRINT(" ,uexc_current_func*2 0x");                    
            UPRINTU16((U16)uexc_current_func*2);                   
   
            dumpTcb(current_task);                                 
            dumpStack(current_task, spCur);
            #endif
         }
      }
   }
   //
   // looks like the current task is to be left running
   //
}                                    

//
// start the timer interrupt routines initially
//
void UEXC_StartTimer(void)
{
   UEXC_RestartTimer();
   INTR_ON();
}

//
// restart the timer interrupt routines
//
void UEXC_RestartTimer(void)
{
   asm("push r16");            
   asm("clr r16");
   asm("out TCCR2,r16");       // Stop Timer2
   
   asm("out TCNT2,r16");       // Clear Timer2
                                                  
   asm("ldi r16,0xbe");        // Set Compare 2 (5ms is 0x13)
   asm("out OCR2,r16");        // ((4MHz/1024)/190) = 50ms timer

   asm("ldi r16,0x0d");
   asm("out TCCR2,r16");       // Start Timer2 with clk/1024

   asm("in  r16,TIMSK");       // Compare 2 Interrupt enable 
   asm("ori r16,1<<OCIE2");     
   asm("out TIMSK,r16");        
   asm("pop r16");

/***
   asm("push r16");
   asm("clr r16");
   asm("out TCCR1B,r16");      // Stop Timer1
   
   asm("out TCNT1H,r16");      // Clear Timer1
   asm("out TCNT1L,r16");
   
   //asm("ldi r16,0x01");
   //asm("out OCR1AH,r16");    // Set Compare A to 
   //asm("ldi r16,0x7c");
   //asm("out OCR1AL,r16");    // ((4MHz/1024)/380) = 100ms timer
   
   asm("ldi r16,0x0e");
   asm("out OCR1AH,r16");      // Set Compare A to ed8 *** DEBUG ***
   asm("ldi r16,0xd8");
   asm("out OCR1AL,r16");     // ((4MHz/1024)/3800) = 1000 ms timer
   
   asm("ldi r16,1<<OCIE1A");
   asm("out TIMSK,r16");      // Compare A Interrupt enable 
   
   asm("ldi r16,0x0D");
   asm("out TCCR1B,r16");     // Start Timer1 with clk/1024
   asm("pop r16");
***/

}
//
// set leds
//
void showLites(Lite_t* lites)
{
   unsigned char* litePtr;

   litePtr = (unsigned char*)lites;
   
   //PPRINT("\r\nshowLites lites 0x"); PPRINTU8(*litePtr);
         
   *LED = ~(*litePtr);
}

/* Create a new process and link it to the task list
 */
int UEXC_CreateTask(void (*func) (void),        // initial start address
                    unsigned char* stack_start, // address of stack for task
                    unsigned int   stack_size,  // 200 recommended
                    unsigned char* msgQ_start,  // address of msg queue for task
                    unsigned int   msgQ_size,   // UEXC_MIN_MSG_QUE_SIZE recommended
                    unsigned int   millsReady,  // mills per T_READY time slice  
                    unsigned char  prior)       // task's priority 0-255 (0 is highest)  
{
	TaskControlBlock *p, *next;
	
	INTR_OFF();

   if (free_list_ptr == 0)
   {
      InitSystem();
   }

	p = free_list_ptr;
	free_list_ptr = free_list_ptr->next;

   //UPRINT("\r\n\r\nUEXC_CreateTask ");

   p->func = (addr_t)func;       // this is a word pointer!!!!

   //UPRINT("p->func*2 0x"); UPRINTU16((U16)p->func*2);

	p->state = T_CREATED;
	p->tid = next_tid++;
   ++numTasks;

   //UPRINT("\r\nUEXC_CreateTask p->tid 0x"); UPRINTU16((U16)p->tid);

	p->ticksTReady = (millsReady+UEXC_CLOCK_TICK_MILLS-1)/UEXC_CLOCK_TICK_MILLS;  

   p->prior = prior;

	/* stack grows from high address to low address
	 */
	p->stack_start = (U16)stack_start;
	
   p->stack_end = (U16)stack_start+stack_size-1;
   
   *(U8*)p->stack_start = 0x55;                  // for overflow check
   ++p->stack_start;
   *(U8*)p->stack_start = 0x55;                     
   ++p->stack_start;
	
   /* align stack here, if needed
	 */
	p->sp = p->stack_end;   
                                                 // set up msg queue
   p->numMsgOnQue  = 0;      
   p->tmoTicks     = 0;      
   p->msgQueGetPtr = (U8*)msgQ_start;     
   p->msgQuePutPtr = (U8*)msgQ_start;     
   p->msgQue1stPtr = p->msgQueGetPtr;      
   p->msgQueEndPtr = p->msgQueGetPtr + msgQ_size;      

	/* create a circular linked list
	 */
	if (current_task == 0)
		p->next = p, current_task = p;
	else
   {
		next = current_task->next, current_task->next = p, p->next = next;
   }

   //dumpTcb(p);       
	INTR_ON();
	return p->tid;
}


//
// Give up timeslice
//
void UEXC_Defer(void)
{

   INTR_OFF();
                                     // return to caller next time slice
   uexc_current_func  = (addr_t)justReturns;
   current_task->func = (addr_t)justReturns;

                                     // push address for reti later 
   asm(" mov  r0,r24");
   asm(" ldi  r24,low((_justReturns)/2)");
   asm(" push r24");
   asm(" ldi  r24,high((_justReturns)/2)");
   asm(" push r24");
   asm(" mov  r24,r0");
                                     // fake a timer interrupt
   //asm(" jmp __output_compare1a_");   // for debug
   asm(" jmp __output_compare2_");
}

//
// dummy routine to execute a return for defer routine
//
void justReturns(void)
{
   return;
}


#ifdef UEXC_DEBUG
//
// dump current task's stack
//
void dumpStack(TaskControlBlock* tcbPtr, unsigned int spNow)
{
   unsigned int numBytes;
   unsigned int i;
   
   if (spNow == 0)
      spNow = tcbPtr->sp;

   numBytes = tcbPtr->stack_end - spNow;
   UPRINT("\r\n\r\ndumpStack tid 0x"); UPRINTU8(tcbPtr->tid);
   UPRINT(" ,numBytes 0x"); UPRINTU16(numBytes);
   UPRINT("\r\n ,spNow 0x"); UPRINTU16(spNow);
   UPRINT(" ,stack_start 0x"); UPRINTU16(tcbPtr->stack_start);
   UPRINT(" ,stack_end 0x"); UPRINTU16(tcbPtr->stack_end);

   for (i=1; i<=numBytes; ++i)                
   {
      UPRINT("\r\n + 0x"); UPRINTU8(i);
      UPRINT(" ,0x"); UPRINTU16((U16)spNow+i);
      UPRINT("-> 0x"); UPRINTU8(*(U8*)(spNow+i));
      
      if (spNow+i > tcbPtr->stack_end || spNow+i <= tcbPtr->stack_start-1) 
      {
         UPRINT("\r\n OUT OF RANGE");
         break;
      }
   }
}
#endif
                                
/* Hog the processor for more cycle time
 */
void UEXC_HogProcessor(void)
{
	current_task->tmoTicks = current_task->ticksTReady;
}

/* kill a process
 */
void UEXC_KillTask(int tid)
{
	TaskControlBlock **qb, *p;

	/* tid == 0 -> null task, never kill that 
	 */
	if (tid == 0)
		return;

	INTR_OFF();
	/* search the circular list until we find a match or gone through
	 * the whole list
	 */
	for (qb = &(current_task)->next; (*qb)->tid != tid && *qb != current_task;
	     qb = &(*qb)->next)
		;
	/* if found, then delete the process
	 */
	if ((*qb)->tid == tid)
	{
		/* If this is the head of the task list, reassign the
		 * head. This has the effect of skipping the next task
		 * for one round of scheduling but should not matter
		 * much.
		 */
		if ((*qb) == current_task)
			current_task = current_task->next;

		/* delink and free the memory
		 */
		p = *qb;
		*qb = (*qb)->next;
		p->next = free_list_ptr;
		free_list_ptr = p;
                              // total running tasks
      --numTasks;
	}
}

/* Internal Functions
 * ==================
 */
                                     
/* Initialize the system.
 */
static void InitSystem(void)
{
	int i;

	/* initialize the free list
	 */
	for (i = 0; i < NUM_TASKS-1; i++)
		tcb[i].next = &tcb[i+1];
	free_list_ptr = &tcb[0];

   for (i=0; i<NUM_TASKS; ++i) 
   {
     attachedTasks[i].timer = 0;                
     attachedTasks[i].ticks = 0;                
   }

	/* null task has tid of 0 
	 */

   //UPRINT("\r\nInitSystem UEXC_CreateTask NullTask() 0x"); 
   //UPRINTU16((U16)NullTask*2);

	UEXC_CreateTask(NullTask,                       // entry point
                   null_task_stack,                // stack 
                   sizeof(null_task_stack),        // stack size
                   0,                              // msgQ 
                   0,                              // msqQ size
                   1,                              // time slice mills
                   255);                           // priority (lowest possible)              
}

/* This is the "return point" of a new task. Reclaim storage
 * and reschdule
 */
void UEXC_KillSelf(void)
{
   INTR_OFF();

   //UPRINT("\r\nUEXC_KillSelf calling UEXC_KillSelf current_task->tid 0x");
   //UPRINTU8(current_task->tid);

   UEXC_KillTask(current_task->tid);

   //UPRINT("\r\nUEXC_KillSelf calling UEXC_Schedule"); 
	UEXC_Schedule();
}

/* Always runnable task. This has the tid of zero
 */
void NullTask(void)
{
	while (1)
   {
      ; // idle indication, could flash a light ??? code me ???
   }
}

/* called by the timer interrupt. See if we need to reschedule
 */
void UEXC_CheckTask(void)
{

   if (current_task == 0)
      return; // no tasks created yet
                                           // watch out for drivers
   if(current_task->tid < numTasks &&
      *((U16*)current_task->stack_start-1) != 0x5555)
   {
      UPRINT("\r\nUEXC_CheckTask stack overflowed tid 0x");
      UPRINTU8(current_task->tid);
   }

   UEXC_RunAttachedTask();      // won't return if task is run 

   UEXC_DecTaskTimers();        // decrement all task timers being run


   //UPRINT("\r\nUEXC_CheckTask tmoTicks 0x");
   //UPRINTU16(current_task->tmoTicks);
   //UPRINT(" ,tid 0x"); UPRINTU8(current_task->tid);
   //UPRINT(" ,func*2 0x"); UPRINTU16(current_task->func*2);
   //UPRINT(" ,justRetAddr 0x"); UPRINTU16((U16)justRetAddr);
   
   //UEXC_DecTimer();
                                   // timer expired or task defered   
	if (current_task->tmoTicks==0 || pcCur==(addr_t)justReturns)
	{
      //
      // validate stack pointer is for current task
      //                
      if (spCur <= current_task->stack_end   && 
          spCur >= current_task->stack_start &&
          current_task->tid < numTasks)
      {
                                     // memorize the state
         uexc_current_sp    = spCur;
         current_task->sp   = spCur;
   
         uexc_current_func  = pcCur;
         current_task->func = pcCur;
         
         //UPRINT("\r\nUEXC_CheckTask uexc_current_sp 0x");
         //UPRINTU16((U16)uexc_current_sp);
   
         //UPRINT("\r\nUEXC_CheckTask uexc_current_func 0x");
         //UPRINTU16((U16)uexc_current_func);
   
         //UPRINT("\r\nUEXC_CheckTask calling UEXC_Schedule"); 
   
         UEXC_Schedule();
      }
      else
      {
         if (current_task->tid < numTasks) 
         {
            UPRINT("\r\n_TimerInterrupt invalid stack, tid 0x"); 
            UPRINTU8((U8)current_task->tid);
            #ifdef UEXC_DEBUG                                 
            UPRINT(" spCur 0x"); UPRINTU16((U16)spCur);                                 //
            UPRINT(" ,pcCur 0x"); UPRINTU16((U16)pcCur);           
            UPRINT(" ,uexc_current_func*2 0x");                    
            UPRINTU16((U16)uexc_current_func*2);                   
   
            dumpTcb(current_task);                                 
            dumpStack(current_task, spCur);
            #endif
         }
      }
	}
}

//
// check for attached task startup time
//
void UEXC_RunAttachedTask(void)
{
	static unsigned char t;
   
   if (pcCur == (addr_t)justReturns)
      return; // ignore fake interrupts from defer

   for (t=0; t<numTasks; ++t)            
   {
     if (attachedTasks[t].ticks > 0)                // attached task ?
     {
        //UPRINT("\r\nUEXC_RunAttachedTask t 0x"); UPRINTU8(t);
        //UPRINT(" ,ticks 0x"); UPRINTU16(attachedTasks[t].ticks);
        //UPRINT(" ,tid 0x");   UPRINTU8(current_task->tid);

        --attachedTasks[t].ticks;                   // yep, time to start it?


        //
        // time to kick off task attached to clock
        //
        if (attachedTasks[t].ticks == 0)
        {                                                                                   // if its not us
           //UPRINT(" ,attachedTasks[t].ticks 0x");
           //UPRINTU16(attachedTasks[t].ticks);

           if (current_task->tid != t)              // some other task running? 
           {
              if (spCur <= current_task->stack_end   && 
                  spCur >  current_task->stack_start &&
                  current_task->tid < numTasks)
              {                                     // check point interrupted task
                 current_task->sp   = spCur;
                 current_task->func = pcCur;
                 
                 //UPRINT(" ,sp 0x");   UPRINTU16(current_task->sp);
                 //UPRINT(" ,func 0x"); UPRINTU16(current_task->func);
                 //dumpTcb(current_task);
                 //dumpStack(current_task, spCur);
              }
           }
           else      
           {
              attachedTasks[t].ticks = attachedTasks[t].timer;
              
              //UPRINT("\r\nUEXC_RunAttachedTask already running tid 0x");
              //UPRINTU8(current_task->tid);
              
              return; // attached task is currently running, try next time
           }

           //
           // time to kick him off
           //
                                                    // set attach clock timout value
           attachedTasks[t].ticks = attachedTasks[t].timer;
                                                    
           current_task = &tcb[t];                  // point to attached task
           
           //UPRINT("\r\nUEXC_RunAttachedTask calling UEXC_ResumeTask tid 0x");
           //UPRINTU8(current_task->tid);

           current_task->state = T_READY;   
           current_task->tmoTicks = current_task->ticksTReady;

           UEXC_ResumeTask();    // won't return, jumps to task
        }
     }
   }
}



//
// decrement sll task times 
//
void UEXC_DecTaskTimers(void)
{
	static unsigned char t;
   
   if (pcCur == (addr_t)justReturns)
      return; // ignore fake interrupts from defer

   for (t=0; t<numTasks; ++t)            
   {
     
     if (tcb[t].tmoTicks != UEXC_NO_TIMEOUT && tcb[t].tmoTicks != 0)
     {
        //UPRINT("\r\nDecTaskTimers t 0x"); UPRINTU8(t);
        //UPRINT(" ,ticks 0x"); UPRINTU16(tcb[t].ticks);
        //UPRINT(" ,tid 0x");   UPRINTU8(current_task->tid);

        --tcb[t].tmoTicks;            // decrement it but leave one tick?
     }
   }
}                            

/* Run a different task. Set the current task as the next one selected in the (circular)
 * list, then set the global variables and call the appropriate asm routines
 * to do the job.
 */
void UEXC_Schedule(void)
{
   static char   t;
   #ifdef HDSL2
   linest_t*     hdsl2_lptr;
   unsigned char lineNum;
   #endif


   if (current_task == 0)
      return;     // nothing to schedule

//
// get next task in circular list
//
   
   INTR_OFF();
   for (t=0; t<numTasks; ++t) 
   {                                           
                                               
      #ifdef PREMPTIVE
      current_task = getHighestPriorTask();   // get highest priority ready task
      #else
      current_task = current_task->next;      // select next task round robin 
      #endif

      //UPRINT("\r\n\r\nUEXC_Schedule trying tid 0x"); //
      //UPRINTU8(current_task->tid);                   //
      //UPRINT(" ,state 0x");                          //
      //UPRINTU8(current_task->state);                 //

                                              // skip attached tasks
      if (attachedTasks[current_task->tid].timer != 0)
      {
         //UPRINT("\r\nUEXC_Schedule found attached tid 0x"); 
         //UPRINTU8(current_task->tid);
         continue;  
      }

      //
      // act on task's current state   
      //   
   	switch (current_task->state)
      {
         case T_READY:                         // task ready to run again
   
            //UPRINT("\r\nUEXC_Schedule T_READY calling UEXC_ResumeTask");
            current_task->tmoTicks = current_task->ticksTReady;
      		UEXC_ResumeTask();    // won't return, jumps to task
            
         case T_CREATED:                       // task ready for 1st run
   	
   		   current_task->state = T_READY;
            current_task->tmoTicks = current_task->ticksTReady;
            //UPRINT("\r\nUEXC_Schedule T_CREATED calling UEXC_StartNewTask");
            UEXC_StartNewTask();  // won't return, jumps to new task
   	
         case T_WF_MSG:                        // task is blocked waiting for msg
      
            //UPRINT("\r\nUEXC_Schedule T_WF_MSG state");
            if (current_task->tmoTicks == 0 || current_task->numMsgOnQue > 0)
            {
               current_task->state = T_READY;
               current_task->tmoTicks = current_task->ticksTReady;
               UEXC_ResumeTask();    // won't return, jumps to task 
            }
            continue; // try next task
   
         case T_WF_EOC_MSG:                    // task is blocked waiting for eoc msg
         
            //UPRINT("\r\nUEXC_Schedule T_WF_EOC_MSG state, tmoTicks 0x");
            //UPRINTU16(current_task->tmoTicks);
            #ifdef HDSL2
            lineNum = current_task->eocLineNum;
            if (lineNum > 0 && lineNum <= L1NumberDslLines)
            {
               hdsl2_lptr = cfg_stat2ptr[lineNum-1];
               if (hdsl2_lptr->eocRcvdMsg > 0)
               {
                  current_task->state = T_READY;
                  current_task->tmoTicks = current_task->ticksTReady;
                  UEXC_ResumeTask();    // won't return, jumps to task 
               }
            }
            #endif              
            if (current_task->tmoTicks == 0 )
            {
               current_task->state = T_READY;
               current_task->tmoTicks = current_task->ticksTReady;
               UEXC_ResumeTask();    // won't return, jumps to task 
            }
            continue;

         case T_WF_SIGNAL:                     // task is blocked waiting for a signal
            
            if (current_task->tmoTicks == 0 || 
                current_task->signal != UEXC_NO_SIGNAL_PRESENT)
            {
               current_task->state = T_READY;
               current_task->tmoTicks = current_task->ticksTReady;
               //UPRINT("\r\nUEXC_Schedule T_WF_SIGNAL calling UEXC_ResumeTask");
               UEXC_ResumeTask();    // won't return, jumps to task 
            }
            continue; // try next task
   
         case T_SLEEPING:                      // task sleeping for a while
            
            //UPRINT("\r\nUEXC_Schedule T_SLEEPING state tmoTicks 0x");
            //UPRINTU16(current_task->tmoTicks);

            //UEXC_DecTimer();
            if (current_task->tmoTicks == 0)
            {
               current_task->state = T_READY;
               current_task->tmoTicks = current_task->ticksTReady;
               UEXC_ResumeTask();    // won't return, jumps to task 
            }
            continue; // try next task
   
         case T_STOPPED:                       // task stopped
   
            //UPRINT("\r\nUEXC_Schedule T_STOPPED state");
            continue;
   
    
         default:                     
            //UPRINT("\r\nUEXC_Schedule unknown state 0x");
            //UPRINTU8(current_task->state);
            //UPRINT(" ,tid 0x");UPRINTU8(current_task->tid);
   
      }
   }
   //UPRINT("\r\nUEXC_Schedule no task to run, can't happen");
}

//
// getHighestPriorTask
//
TaskControlBlock* getHighestPriorTask (void)
{
   TaskControlBlock *bestTask, *nextTask;
   static char t;
//
// scan all tasks for  highest priority
//
   
   bestTask = 0;     
   for (t=0; t<numTasks; ++t) 
   {                                      // get next task in list     
      nextTask = current_task->next;  
                                          // skip attached tasks
      if (attachedTasks[current_task->tid].timer != 0)
         continue;  
      
      if (bestTask == 0) 
         bestTask = nextTask;

      if (nextTask->prior < bestTask->prior) 
      {
         bestTask = nextTask;         // save highest priority task

         //UPRINT("\r\ngetHighestPriorTask bestTask->tid 0x");   UPRINTU8(bestTask->tid);
         //UPRINT("\r\ngetHighestPriorTask bestTask->prior 0x"); UPRINTU8(bestTask->prior);
      }

   }
   //if (bestTask == 0) 
   //   UPRINT("\r\ngetHighestPriorTask bestTask is 0, can't happen");
   return bestTask;
}

//
// decrement tasks timer
//
//void UEXC_DecTimer(void)
//{                                
//                                // ignore fake interrupts from UEXC_Defer   
//   if (pcCur == (addr_t)justReturns)
//      return;
//                                // dec timer if running
//   if (current_task->tmoTicks != UEXC_NO_TIMEOUT && current_task->tmoTicks != 0)
//   {
//      --current_task->tmoTicks;
//      //UPRINT("\r\nUEXC_DecTimer tmoTicks 0x");
//      //UPRINTU16(current_task->tmoTicks);
//      //UPRINT(" ,tid 0x");
//      //UPRINTU16(current_task->tid);
//   }
//}

//
// start a new task
//
void UEXC_StartNewTask(void)
{
   static unsigned char* spTmp;
   static unsigned char* taskSpHi;
   static unsigned char* taskSpLo; 
   static char i;

   INTR_OFF();
//
// set up new task's initial stack
//
   spTmp = (U8*)current_task->sp;
                                         // in case task returns
   *spTmp = (addr_t)UEXC_KillSelf & 0xff;         
   --spTmp;
   
   *spTmp = (addr_t)UEXC_KillSelf >> 8;
   --spTmp;

   
//
// create initial stack frame
//
   
   *spTmp = (addr_t)current_task->func & 0xff;  // 'reti' address   
   --spTmp;

   *spTmp = (addr_t)current_task->func >> 8;   
   --spTmp;
   
   *spTmp = 0;                          // __zero_reg__
   --spTmp;

   *spTmp = 0;                          // __tmp_reg__
   --spTmp;
   
   *spTmp = 0x00;                       // SREG, interrupts off, flags clear
   --spTmp;

   for (i=0; i<=27; ++i)                // r0 to r27 
   {
      *spTmp = 0;
      --spTmp;
   }
   taskSpLo = spTmp;                    // save pointer to new tasks 
   --spTmp;                                // stack pointer low byte

   taskSpHi = spTmp;                    // save pointer to new tasks 
   --spTmp;                                // stack pointer high byte
   
   for (i=0; i<2; ++i)                  // r30 to r31 
   {
      *spTmp = 0;
      --spTmp;
   }

                                        // save the initial stack pointer
   *taskSpLo = (U16)spTmp & 0xff;
   *taskSpHi = (U16)spTmp >> 8;
                                        // adjust to look like real
                                        // interrupt stack frame 
   current_task->sp = (U16)spTmp;

   //UPRINT("\r\n\r\nUEXC_StartNewTask sp 0x"); UPRINTU16(current_task->sp);
   //dumpTcb(current_task);
   //dumpStack(current_task, (U16)spTmp);
   
   UEXC_ResumeTask(); //doesn't return 
}


//
// resume task execution
//

void UEXC_ResumeTask(void)
{
   addr_t           spTmp;        // temp stack pointer
   unsigned char    dmy;          // don't want stuff on stack
   

   INTR_OFF();
                                  // sp before pop's 
   spTmp  = current_task->sp;
   
   uexc_current_sp   = current_task->sp;

   uexc_current_func = (addr_t)current_task->func;

   //UPRINT("\r\n ,spTmp 0x");   
   //UPRINTU16((U16)spTmp);                        
                                                 
   //UPRINT(" ,uexc_current_sp 0x");               
   //UPRINTU16((U16)uexc_current_sp);              
                                                 
   //dumpTcb(current_task);                        
   //dumpStack(current_task, spTmp); 
   
//
// restore stack frame created by timer interrupt handler
//
// **** clears interrupt flag when SREG is restored ****
//
   
   asm volatile (
      "mov r28,%A1                    "  "\n\t"
      "mov r29,%B1                    "  "\n\t"
   	"out __SP_L__,r28               "  "\n\t"
   	"out __SP_H__,r29               "  "\n\t"
   	"pop r31                        "  "\n\t"
   	"pop r30                        "  "\n\t"
   	"pop r29                        "  "\n\t"
   	"pop r28                        "  "\n\t"
   	"pop r27                        "  "\n\t"
   	"pop r26                        "  "\n\t"
   	"pop r25                        "  "\n\t"
   	"pop r24                        "  "\n\t"
   	"pop r23                        "  "\n\t"
   	"pop r22                        "  "\n\t"
   	"pop r21                        "  "\n\t"
   	"pop r20                        "  "\n\t"
   	"pop r19                        "  "\n\t"
   	"pop r18                        "  "\n\t"
   	"pop r17                        "  "\n\t"
   	"pop r16                        "  "\n\t"
   	"pop r15                        "  "\n\t"
   	"pop r14                        "  "\n\t"
   	"pop r13                        "  "\n\t"
   	"pop r12                        "  "\n\t"
   	"pop r11                        "  "\n\t"
   	"pop r10                        "  "\n\t"
   	"pop r9                         "  "\n\t"
   	"pop r8                         "  "\n\t"
   	"pop r7                         "  "\n\t"
   	"pop r6                         "  "\n\t"
   	"pop r5                         "  "\n\t"
   	"pop r4                         "  "\n\t"
   	"pop r3                         "  "\n\t"
   	"pop r2                         "  "\n\t"
   	"pop r1                         "  "\n\t"
   	"pop r0                         "  "\n\t"
   	"pop __tmp_reg__                "  "\n\t"
      "lsl __tmp_reg__                "  "\n\t"
      "lsr __tmp_reg__                "  "\n\t"
   	"out __SREG__,__tmp_reg__       "  "\n\t"
   	"pop __tmp_reg__                "  "\n\t"
   	"pop __zero_reg__               "  "\n\t"
  		: "=r" (dmy) 			
  		: "r"  (spTmp)	
  		: "r30", "r31"
   );
   
      asm("reti");                 // turns on interrupts

}

//
// send msg to task  (callable by drivers)
//
char UEXC_SendMsg( unsigned char* msgBuf,    // user buffer holding msg
                   unsigned char  msgSize,   // byte length of msg to send
                   unsigned char  fromTid,   // task id sending msg (if any)
                   unsigned char  destTid)   // task id to sent to
{
   static TaskControlBlock* destTcbPtr;
   static unsigned char* startPutPtr;      
   static unsigned char  i;      

   if (destTid >= numTasks)
   {
      //UPRINT("\r\nUEXC_SendMsg destTid 0x"); UPRINTU8(destTid);
      //UPRINT(" invalid, numTasks 0x"); UPRINTU8(numTasks);
      return 0; // invalid tid
   }

   //UPRINT("\r\nUEXC_SendMsg queueing 0x"); UPRINTU8(msgSize);
   //UPRINT(" to destTid 0x"); UPRINTU8(destTid);
                                                 // tcb ptr
   destTcbPtr = &tcb[destTid];

   if (destTcbPtr->msgQue1stPtr == 0)          
   {
      //UPRINT("\r\nUEXC_SendMsg no msg queue");
      return 0; // no msg queue
   }
                                              // save msg starting point 
   startPutPtr = destTcbPtr->msgQuePutPtr;
                                              // task id sending msg
   *destTcbPtr->msgQuePutPtr = fromTid;  
   
   if (incMsgQuePutPtr(destTcbPtr) == 0) 
   {
      destTcbPtr->msgQuePutPtr = startPutPtr;
      //UPRINT("\r\nUEXC_SendMsg queue full saving fromTid");   
      return 0;  // que full
   }
                                              // number of bytes
   *destTcbPtr->msgQuePutPtr = msgSize;

   if (incMsgQuePutPtr(destTcbPtr) == 0) 
   {
      destTcbPtr->msgQuePutPtr = startPutPtr;
      //UPRINT("\r\nUEXC_SendMsg queue full saving msgSize");   
      return 0;  // que full
   }
                                              // add msg data to xmt que
   for (i=0; i<msgSize; ++i)
   { 
      *destTcbPtr->msgQuePutPtr = msgBuf[i];

      if (incMsgQuePutPtr(destTcbPtr) == 0)
      {
         destTcbPtr->msgQuePutPtr = startPutPtr;
         //UPRINT("\r\nUEXC_SendMsg queue full saving data byte");      
         return 0;  // que full
      }
   }

                          // tell destination tasks a msg is on its queue
   ++destTcbPtr->numMsgOnQue;

   return 1;   
}

//
// read msg, blocks with timeout
//
char UEXC_ReadMsg( unsigned char* msgBuf,    // user buffer to recv msg
                   unsigned char  bufSize,   // byte size of buffer
                   unsigned char* xmtTid,    // task id that sent msg
                   unsigned char* msgSize,   // byte size of msg rcvd
                   unsigned int   tmoMills)  // max millisec to wait
{
   //static unsigned char retVal;
   
   if (current_task->msgQue1stPtr == 0) 
      return 0; // no msg queue
   
   //UPRINT("\r\nUEXC_ReadMsg numMsgOnQue 0x");
   //UPRINTU8(current_task->numMsgOnQue);
   //UPRINT(" , tid 0x"); UPRINTU8(current_task->tid);
   //UPRINT(" , &msgBuf 0x"); UPRINTU16((U16)msgBuf);
   //UPRINT(" , &xmtTid 0x"); UPRINTU16((U16)xmtTid);
   //UPRINT(" , &msgSize 0x"); UPRINTU16((U16)msgSize);

                                      // check msg already available
   if (current_task->numMsgOnQue > 0)
   {
      
      //UPRINT("\r\nUEXC_ReadMsg msg already avail");
      return UEXC_unqueMsgRcvd(msgBuf, bufSize, xmtTid, msgSize, tmoMills);
      
      //if (retVal == 0)
      //   UPRINT("\r\nUEXC_ReadMsg UEXC_unqueMsgRcvd failed");
      //return retVal;
   }
                                     
                                      // gotta wait for a msg or tmo
   //UPRINT("\r\nUEXC_ReadMsg gotta wait for a msg");
   
   if (tmoMills == 0 || tmoMills == UEXC_NO_TIMEOUT) 
      current_task->tmoTicks = UEXC_NO_TIMEOUT; // never time out
   else
      current_task->tmoTicks = (tmoMills+UEXC_CLOCK_TICK_MILLS-1)/UEXC_CLOCK_TICK_MILLS;

   //UPRINT("\r\nUEXC_ReadMsg tmoTicks 0x");     
   //UPRINTU16(current_task->tmoTicks);

   //UPRINT("\r\nUEXC_ReadMsg calling UEXC_Defer");

   current_task->state = T_WF_MSG;
   UEXC_Defer(); // returns on tmo or msg  

   //UPRINT("\r\nUEXC_ReadMsg numMsgOnQue 0x");
   //UPRINTU8(current_task->numMsgOnQue);
   //UPRINT(" , tid 0x"); UPRINTU8(current_task->tid);
   
   if (current_task->numMsgOnQue > 0 )
   {
      //dumpMsgQue(current_task);
      return UEXC_unqueMsgRcvd(msgBuf, bufSize, xmtTid, msgSize, tmoMills);
      //if (retVal == 1)
      //{
      //   current_task->tmoTicks = current_task->ticksTReady;
      //   current_task->state = T_READY;
      //   return 1;
      //}
      //else
      //   UPRINT("\r\nUEXC_ReadMsg UEXC_unqueMsgRcvd failed");
   } 
   return 0;  // timed out
}


//
// pull next msg off task msg queue
//
char UEXC_unqueMsgRcvd( 
                   unsigned char* msgBuf,    // user buffer to recv msg
                   unsigned char  bufSize,   // byte size of buffer
                   unsigned char* xmtTid,    // task id that sent msg
                   unsigned char* msgSize,   // byte size of msg rcvd
                   unsigned int   tmoMills)  // max millisec to wait
{
   unsigned char* startGetPtr;
   unsigned char i;

   //UPRINT("\r\nunqueMsgRcvd ");
   //UPRINT(" , &msgBuf 0x"); UPRINTU16((U16)msgBuf);
   //UPRINT(" , &xmtTid 0x"); UPRINTU16((U16)xmtTid);
   //UPRINT(" , &msgSize 0x"); UPRINTU16((U16)msgSize);

   startGetPtr = current_task->msgQueGetPtr;
                                            // givem sending tid
   *xmtTid = *current_task->msgQueGetPtr;
   if (incMsgQueGetPtr(current_task) == 0)
   {
      //UPRINT("\r\nunqueMsgRcvd msgQueData corrupt count");
      //UPRINT(" ,tid 0x"); 
      //UPRINTU8(current_task->tid);
      //dumpMsgQue(current_task);
      return 0;  // no more data
   }
                                           // givem msg byte length
   *msgSize = *current_task->msgQueGetPtr;
   
   //UPRINT("\r\nUEXC_unqueMsgRcvd *msgSize 0x"); 
   //UPRINTU8(*msgSize);  

   if (incMsgQueGetPtr(current_task) == 0)
   {
      //UPRINT("\r\nUEXC_unqueMsgRcvd msgQueData corrupt count , *msgSize 0x");
      //UPRINTU8(*msgSize);
      //UPRINT(" ,tid 0x"); UPRINTU8(current_task->tid);
      //dumpMsgQue(current_task);
      return 0;  // no more data
   }

   for (i=0; i<*msgSize; ++i)
   {                                          // givem data bytes
      if (i < bufSize) 
         msgBuf[i] = *current_task->msgQueGetPtr;

      //UPRINT("\r\nUEXC_unqueMsgRcvd msgQueGetPtr 0x"); 
      //UPRINTU16((U16)current_task->msgQueGetPtr);
      //UPRINT(" ,msgBuf[");UPRINTU8(i);
      //UPRINT("]: 0x");UPRINTU8(msgBuf[i]);

      if (incMsgQueGetPtr(current_task) == 0)
      {
         //UPRINT("\r\nUEXC_unqueMsgRcvd msgQueData corrupt data, *msgSize 0x");
         //UPRINTU8(*msgSize);
         //UPRINT(" ,tid 0x"); UPRINTU8(current_task->tid);
         //dumpMsgQue(current_task);   
         return 0;  // no more data
      }
   }
                                              // got msg
   --current_task->numMsgOnQue;

   //UPRINT("\r\nUEXC_unqueMsgRcvd left with numMsgOnQue 0x");
   //UPRINTU8(current_task->numMsgOnQue);
   //UPRINT(" ,*msgSize 0x"); UPRINTU8(*msgSize);
   //UPRINT(" ,bufSize 0x");  UPRINTU8(bufSize);

   if (*msgSize > bufSize) 
      return 0;
   return 1;
}

#ifdef HDSL2
//
// read EOC msg, blocks with timeout (no spin loops)
//
char UEXC_ReadEocMsg( unsigned char* msgBuf,   // user buff to recv msg
                      unsigned char  bufSize,  // bytes size of user buffer
                      unsigned char  lineNum,  // hdsl2 line number 1 to N
                      unsigned char* msgSize,  // byte size of msg rcvd
                      unsigned int   tmoMills) // max millisec to wait
{
   linest_t* irq_lptr;

   //UPRINT("\r\nUEXC_ReadEocMsg tmoMills 0x"); UPRINTU16(tmoMills);
   //UPRINT(" ,lineNum: 0x"); UPRINTU8(lineNum);

                                               // validate hdsl2 line #
   if (lineNum < 1 || lineNum > L1NumberDslLines) 
      return 0;

   irq_lptr = cfg_stat2ptr[lineNum-1];         // set line status table ptr
                                               
   //UPRINT("\r\n ,eocRcvdMsg 0x"); UPRINTU8(irq_lptr->eocRcvdMsg);

   if (irq_lptr->eocRcvdMsg > 0)               // got msgs already in que?
      return unqueEocMsgRcvd(irq_lptr, msgBuf, sizeof(msgBuf), msgSize); // yep

//
// gonna havta wait for the eoc msg to show up
//
   
   if (tmoMills == 0 || tmoMills == UEXC_NO_TIMEOUT) 
      current_task->tmoTicks = UEXC_NO_TIMEOUT; // never time out
   else
      current_task->tmoTicks = (tmoMills+UEXC_CLOCK_TICK_MILLS-1)/UEXC_CLOCK_TICK_MILLS;
   
   //UPRINT(" ,tmoTicks 0x"); UPRINTU8(current_task->tmoTicks);

   current_task->state    = T_WF_EOC_MSG;

   UEXC_Defer(); // return after tmo or eoc msg avail
      
   //UPRINT("\r\nUEXC_ReadEocMsg after UEXC_Defer tmoTicks 0x"); 
   //UPRINTU16(current_task->tmoTicks);
   //UPRINT(" ,eocRcvdMsg 0x"); UPRINTU8(irq_lptr->eocRcvdMsg);

   if (irq_lptr->eocRcvdMsg > 0)             // givem msg just received
      return unqueEocMsgRcvd(irq_lptr, msgBuf, sizeof(msgBuf), msgSize);
   else
   {
      ++irq_lptr->eocRcvTmos;   // eoc msg timeout
      //UPRINT("\r\nUEXC_ReadEocMsg timed out after defer");
   }
   return 0;
}
#endif


#ifdef UEXC_DEBUG
//
// debug dump a msg 
//
void dumpMsg(TaskControlBlock* tcbPtr, unsigned char* rcvStartPtr, 
               unsigned char* rcvPutPtr)
{
   UPRINT("\r\ndumpMsg rcvStartPtr: 0x"); UPRINTU16((U16)rcvStartPtr);
   UPRINT(" ,rcvPutPtr: 0x"); UPRINTU16((U16)rcvPutPtr);

   while (rcvStartPtr != rcvPutPtr) 
   {                                       
      UPRINTU8(*rcvStartPtr); UPRINT(" "); 

      ++rcvStartPtr;
      if (rcvStartPtr >= tcbPtr->msgQueEndPtr) 
         rcvStartPtr = tcbPtr->msgQue1stPtr;
   }
}
#endif

//
// check if msgs available
//
char UEXC_ifMsgAvail()
{
   return current_task->numMsgOnQue;
}


//
// wait for signal from another task
//
char UEXC_WaitSignal(unsigned int   tmoMills, // time to wait, forever (0)
                     unsigned char* sigVal,   // signal number received
                     unsigned char* fromTid)  // task id that sent signal
{
                                     // check signal already available

   //PPRINT("\r\nUEXC_WaitSignal pending signal 0x"); PPRINTU8(current_task->signal);

   if (current_task->signal != UEXC_NO_SIGNAL_PRESENT)
   {
      *sigVal  = current_task->signal;
      *fromTid = current_task->tid;
      current_task->signal = UEXC_NO_SIGNAL_PRESENT;    // clear
      return 1;
   } 
                                     // no signal yet, wait a while
   current_task->state = T_WF_SIGNAL;

   if (tmoMills == 0 || tmoMills == UEXC_NO_TIMEOUT) 
      current_task->tmoTicks = UEXC_NO_TIMEOUT; // never time out
   else
      current_task->tmoTicks = (tmoMills+UEXC_CLOCK_TICK_MILLS-1)/UEXC_CLOCK_TICK_MILLS;
   
   //PPRINT("\r\nUEXC_WaitSignal UEXC_Defer'ing, tid"); PPRINTU8(current_task->tid);
   
   UEXC_Defer(); // returns on tmo or signal  
   
   //PPRINT("\r\nUEXC_WaitSignal UEXC_Defer signal 0x"); PPRINTU8(current_task->signal);
   
   if (current_task->signal != UEXC_NO_SIGNAL_PRESENT )
   {                                 // got signal
      *sigVal  = current_task->signal;   
      *fromTid = current_task->tid;
      current_task->signal = UEXC_NO_SIGNAL_PRESENT;    // clear
      return 1;
   }

   if (current_task->tmoTicks == 0)
   {
      current_task->signal = UEXC_NO_SIGNAL_PRESENT;    // clear
      return 0; // timed out
   }

   //UPRINT("\r\nUEXC_WaitSignal no signal and no tmo, can't happen");
   current_task->signal = UEXC_NO_SIGNAL_PRESENT;    // clear
   return 0;
}

//
// send signal to task
//
char UEXC_SendSignal( unsigned char sigVal,   // signal number to send
                      unsigned char destTid)  // task id to recv signal
{
   if (destTid >= numTasks || sigVal == UEXC_NO_SIGNAL_PRESENT)
      return 0;
                                             // tell em who wants signal
   tcb[destTid].signal  = sigVal;
   tcb[destTid].sigTid  = current_task->tid;

   //UPRINT("\r\nUEXC_SendSignal destTid 0x"); UPRINTU8(destTid);
   //UPRINT(", sigVal 0x"); UPRINTU8(sigVal);
   return 1;
}

//
// task wants to sleep for a time
//
char UEXC_Sleep(unsigned int seconds,    // number of seconds to sleep
                unsigned int mills)      // number if milliseconds to sleep
{

   if (seconds > UEXC_MAX_SLEEP_SECONDS && seconds != UEXC_SLEEP_ATTACHED)
      return 0;
   

   if (seconds != UEXC_SLEEP_ATTACHED)   // we are attachd to clock, it runs timer
      current_task->tmoTicks = 
         (seconds*1000 +mills +UEXC_CLOCK_TICK_MILLS-1)/ UEXC_CLOCK_TICK_MILLS;
   
   //UPRINT("\r\nUEXC_Sleep tmoTicks 0x"); 
   //UPRINTU16(current_task->tmoTicks);
   //UPRINT(" tid 0x"); UPRINTU8(current_task->tid);

   if (seconds == UEXC_SLEEP_ATTACHED)
      current_task->state = T_READY;
   else
      current_task->state = T_SLEEPING;
   UEXC_Defer(); // returns on tmo 
   
   return 1;
}

//
// task gets run periodically
//
char UEXC_Attach(unsigned int seconds,          // number of seconds to sleep
                 unsigned int mills)            // number if milliseconds to sleep
{

   //UPRINT("\r\nUEXC_Attach seconds 0x"); UPRINTU8(seconds);
   //UPRINT(" ,mills 0x"); UPRINTU8(mills);
   //UPRINT("\r\n ,tid 0x"); UPRINTU8(current_task->tid);

   attachedTasks[current_task->tid].timer = (U16)
       ((U32)seconds*1000 + (U32)mills)/UEXC_CLOCK_TICK_MILLS;

   //UPRINT(" ,timer 0x"); UPRINTU8(attachedTasks[current_task->tid].timer);

   attachedTasks[current_task->tid].ticks = attachedTasks[current_task->tid].timer;

   //UPRINT(" ,ticks 0x"); UPRINTU8(attachedTasks[current_task->tid].ticks);
   
   return 1;
}


//
// get task scheduler's clock ticks
//                                              
unsigned int UEXC_GetClockTicks()              
{
   return intrTicks;  // current ticks, UEXC_CLOCK_TICK_MILLS each
}


#ifdef UEXC_DEBUG
//
// debug dump of task control block
//
void dumpTcb(TaskControlBlock* tcbPtr)
{

   UPRINT("\r\n\r\ndumpTcb tid 0x"); UPRINTU8(tcbPtr->tid);
	UPRINT(" ,next 0x"); UPRINTU16((U16)tcbPtr->next);
	UPRINT(" ,prior 0x"); UPRINTU8(tcbPtr->prior);
	UPRINT(" ,state 0x"); UPRINTU8(tcbPtr->state);
	UPRINT(" ,ticksTReady 0x"); UPRINTU16(tcbPtr->ticksTReady);
	UPRINT(" ,func*2 0x"); UPRINTU16((addr_t)tcbPtr->func*2);
	UPRINT("\r\n ,stack_start 0x"); UPRINTU16(tcbPtr->stack_start);
	UPRINT(" ,stack_end 0x"); UPRINTU16(tcbPtr->stack_end);
	UPRINT(" ,sp 0x"); UPRINTU16(tcbPtr->sp);
	UPRINT("\r\n ,tmoTicks 0x"); UPRINTU16(tcbPtr->tmoTicks);
	UPRINT(" ,numMsgOnQue 0x"); UPRINTU8(tcbPtr->numMsgOnQue);
	UPRINT(" ,msgQueGetPtr 0x"); UPRINTU16((U16)tcbPtr->msgQueGetPtr);
	UPRINT(" ,msgQuePutPtr 0x"); UPRINTU16((U16)tcbPtr->msgQuePutPtr);
	UPRINT("\r\n ,msgQue1stPtr 0x"); UPRINTU16((U16)tcbPtr->msgQue1stPtr);
	UPRINT(" ,msgQueEndPtr 0x"); UPRINTU16((U16)tcbPtr->msgQueEndPtr);
	UPRINT(" ,signal 0x"); UPRINTU8(tcbPtr->signal);
	UPRINT(" ,sigTid 0x"); UPRINTU8(tcbPtr->sigTid);  
}
#endif
//
// increment task's msg que put pointer
//
int incMsgQuePutPtr(TaskControlBlock* tcbPtr)
{
   unsigned char* nextPutPtr;
   
   nextPutPtr = tcbPtr->msgQuePutPtr + 1;       // try to bump it

   if (nextPutPtr >= tcbPtr->msgQueEndPtr) 
     nextPutPtr = tcbPtr->msgQue1stPtr;         // wrapped

                                                // check for room  
   if (nextPutPtr == tcbPtr->msgQueGetPtr)
   {                                                           
      //UPRINT("\r\nmsgQueData FULL tid 0x"); 
      //UPRINTU8(tcbPtr->tid);
      return 0;                                 // no room
   }

   tcbPtr->msgQuePutPtr = nextPutPtr;           // update ptr

   //UPRINT("\r\nincMsgQuePutPtr msgQuePutPtr 0x");
   //UPRINTU16((U16)tcbPtr->msgQuePutPtr);

   return 1;
}


//
// increment task's msg que get pointer
//
int incMsgQueGetPtr(TaskControlBlock *tcbPtr)
{
   unsigned char* nextGetPtr;
                                                // check que empty
   if (tcbPtr->msgQueGetPtr == tcbPtr->msgQuePutPtr)
      return 0;  // nothing to get 
   
   nextGetPtr = tcbPtr->msgQueGetPtr + 1;       // try to bump it

   if (nextGetPtr >= tcbPtr->msgQueEndPtr) 
     nextGetPtr = tcbPtr->msgQue1stPtr;         // wrapped

   tcbPtr->msgQueGetPtr = nextGetPtr;           // update ptr

   //UPRINT("\r\nincMsgQueGetPtr msgQueGetPtr 0x");
   //UPRINTU16((U16)tcbPtr->msgQueGetPtr);

   return 1;
}


#ifdef UEXC_DEBUG
//
// dump task's msg que contents
//
void dumpMsgQue(TaskControlBlock *tcbPtr)
{
   unsigned char* nextGetPtr;

   UPRINT("\r\ndumpMsgQue msgQueGetPtr 0x");
   UPRINTU16((U16)tcbPtr->msgQueGetPtr); 
   
   UPRINT(" ,msgQuePutPtr 0x");
   UPRINTU16((U16)tcbPtr->msgQuePutPtr); 
                                                // check que empty
   if (tcbPtr->msgQueGetPtr == tcbPtr->msgQuePutPtr)
   {
      return;  // nothing to get 
   }

   nextGetPtr = tcbPtr->msgQueGetPtr;   
   while (nextGetPtr != tcbPtr->msgQuePutPtr) 
   {
      UPRINT("\r\n nextGetPtr 0x");
      UPRINTU16((U16)nextGetPtr);
      UPRINT(" -> 0x"); UPRINTU8(*nextGetPtr);
      
      ++nextGetPtr;       // try to bump it
   
      if (nextGetPtr >= tcbPtr->msgQueEndPtr) 
        nextGetPtr = tcbPtr->msgQue1stPtr;         // wrapped
   }
}
#endif


#ifdef TRACE_TASKS
//
// dump task trace buffer 
//
void dumpTaskTraces()
{
   unsigned int i;
   
   UPRINT("\r\ndumpTaskTraces");

   for (i=0; i<numTraces; ++i) 
   {
      UPRINT("\r\n tick 0x");  UPRINTU16(traceBuf[i].tick);
      UPRINT(" ,pc 0x");       UPRINTU16(traceBuf[i].pc);
      UPRINT(" ,sp 0x");       UPRINTU16(traceBuf[i].sp);
      UPRINT(" ,tid 0x");      UPRINTU8(traceBuf[i].tid);
      UPRINT(" ,state 0x");    UPRINTU8(traceBuf[i].state);
      UPRINT(" ,tmoTicks 0x"); UPRINTU8(traceBuf[i].tmoTicks);
   }
}
#endif

#ifndef HDSL2
//
// delay given usec
//
void spt_micro_delay(unsigned int time)
{

   volatile unsigned int t;

   for (t=0; t<time/2; t+=3);
}


//
// delay given msec
//
unsigned int msec(unsigned int time)
{

   volatile unsigned int i;

   for (i=0; i<time; ++i)
      spt_micro_delay(1000);

   return 1;
} 
#endif
