#ifndef _UEXEC_H
#define _UEXEC_H


//
// definition of led lights
//

#define UEXC_HEART_BEAT  // remove this to get rid of lights

typedef struct Lite_t
{
   #ifdef HDSL2
   unsigned hdsl2     :5;             // hdsl2 state
   unsigned tid       :3;             // interrupted task id, if any 
   #else // not hdsl2 driver
   unsigned spare     :4;             // spare lights
   unsigned tid       :4;             // interrupted task id, if any
   #endif
} Lite_t;

extern Lite_t lites;


/* Number of tasks, change this to suit your need
 */
#ifndef NUM_TASKS
#define NUM_TASKS	7
#endif
                                             // create a new task
int UEXC_CreateTask( void (*func) (void),       // initial start address
                     unsigned char* stack_start,// address of stack buffer
                     unsigned int   stack_size, // size of stack buffer (200 recommended)
                     unsigned char* msgQ_start, // address of msg que (optional)
                     unsigned int   msgQ_size,  // size of msg que    (optional)
                     unsigned int   mills,      // min mills to run if not blocked
                     unsigned char  prior);     // task's priority 0-255 (0 is highest)  

                                             // startup task scheduler
void UEXC_StartScheduler(void);

                                             // give up current time slice
void UEXC_Defer(void);                          // task runs next time slice

                                             // increase current time slice length
void UEXC_HogProcessor(void);
                                             // take task off task schedule list
void UEXC_KillTask(int tid);

                                             // get msg from task in this processor
char UEXC_ReadMsg( unsigned char* msgBuf,       // user buffer to recv msg
                   unsigned char  bufSize,      // byte size of buffer
                   unsigned char* xmtTid,       // task id that sent msg (if any)
                   unsigned char* msgSize,      // byte size of msg rcvd
                   unsigned int   tmoMills);    // max millisec to wait

                                             // send msg to task in this processor
char UEXC_SendMsg( unsigned char* msgBuf,      // user buffer holding msg
                   unsigned char bufSize,      // byte length of msg to send
                   unsigned char fromTid,      // task id sending msg (if any)
                   unsigned char destTid);     // task id to sent to 0 to N
                                              
                                             // check to see if msgs available
char UEXC_ifMsgAvail();                        // from task in this processor)

                                             // get msg from task in another processor
                                               // (*** doesn't use spin loop ***)
char UEXC_ReadEocMsg( unsigned char* msgBuf,   // user buff to recv msg
                      unsigned char  bufSize,  // bytes size of user buffer
                      unsigned char  lineNum,  // hdsl2 line number 1 to N
                      unsigned char* msgSize,  // byte size of msg rcvd
                      unsigned int   tmoMills);// max millisec to wait

                                             // wait for a signal to arrive
                                               // from task in this processor
char UEXC_WaitSignal( unsigned int   tmoMills,  // time to wait, forever (0)
                      unsigned char* sigVal,   // signal number received
                      unsigned char* fromTid); // task id that sent signal

                                             // send signal to task in this procssor
char UEXC_SendSignal( unsigned char sigVal,    // signal number to send
                      unsigned char destTid);  // task id to recv signal

                                             // suspend task for time specified
char UEXC_Sleep(unsigned int seconds,          // number of seconds to sleep
                unsigned int mills);           // number if milliseconds to sleep

                                             // task gets run periodically
char UEXC_Attach(unsigned int seconds,          // number of seconds to sleep
                 unsigned int mills);           // number of milliseconds to sleep

                                             // get task scheduler's clock ticks 
unsigned int UEXC_GetClockTicks();              // current ticks, UEXC_CLOCK_TICK_MILLS each
                                  

/* Minimum stack size for a task function that does not use "any" stack space.
 * This should be the interrupt stack frame, plus some fudge factor for
 * frame pointer, a single function call, etc.
 * For HC11/12/16, the interrupt frame is 10/12 bytes.
 */
#define UEXC_MIN_STACK_SIZE	(200)    

#define UEXC_INTR_STACK_SIZE  (37)    // includes 'reti' address

#define DEFAULT_TICKS_READY   (2)
#define UEXC_CLOCK_TICK_MILLS (50)
#define UEXC_MIN_MSG_QUE_SIZE (100)

#define UEXC_DEFAULT_TIME_SLICE (UEXC_CLOCK_TICK_MILLS*2)
                                                            
#define PREEMPTIVE  // remove this to get round robin task scheduling 
#define UEXC_LOWEST_PRIOR_TASK 255

//#define TRACE_TASKS // remove this to disable task tracing
#ifdef TRACE_TASKS
#define MAX_TRACE_IDX 20
#else
#define MAX_TRACE_IDX 1
#endif

#define UEXC_NO_TIMEOUT         0xffff  
#define UEXC_SLEEP_ATTACHED     0xfffe  
#define UEXC_MAX_SLEEP_SECONDS  65
#define UEXC_NO_SIGNAL_PRESENT  0  

//#define HDSL2              // define for tasking and hdsl2 driver 
#define UEXC_DEBUG       // define for task debug

#define INTR_OFF()	 asm(" cli");
#define INTR_ON()     asm(" sei");


/* debug print Macros */

//#define UPRINT(string)   UART_PrintfStr(string)
//#define UEOL()           UART_PrintfEndOfLine()
//#define UPRINTU8(data)   UART_PrintfU8(data)
//#define UPRINTU16(data)  UART_PrintfU16(data)
//#define UPRINTU32(data)  UART_PrintfU32(data)

#define UPRINT(string)      // remove debug for uexec
#define UEOL()
#define UPRINTU8(data)          
#define UPRINTU16(data)          
#define UPRINTU32(data)          

#endif
